import random
import re
from typing import Any

from pytest import mark, param, raises, warns

from omegaconf import OmegaConf, Resolver
from omegaconf.nodes import InterpolationResultNode
from tests.interpolation import dereference_node


def test_register_resolver_twice_error(restore_resolvers: Any) -> None:
    def foo(_: Any) -> int:
        return 10

    OmegaConf.register_new_resolver("foo", foo)
    with raises(ValueError, match=re.escape("resolver 'foo' is already registered")):
        OmegaConf.register_new_resolver("foo", foo)


def test_register_resolver_twice_error_legacy(restore_resolvers: Any) -> None:
    def foo() -> int:
        return 10

    OmegaConf.legacy_register_resolver("foo", foo)
    with raises(AssertionError):
        OmegaConf.legacy_register_resolver("foo", lambda: 10)


def test_register_resolver_twice_error_legacy_and_regular(
    restore_resolvers: Any,
) -> None:
    def foo() -> int:
        return 10

    OmegaConf.legacy_register_resolver("foo", foo)
    with raises(ValueError):
        OmegaConf.register_new_resolver("foo", foo)


def test_register_resolver_error_non_callable(restore_resolvers: Any) -> None:
    with raises(TypeError, match=re.escape("resolver must be callable")):
        OmegaConf.register_new_resolver("foo", 0)  # type: ignore


def test_register_resolver_error_empty_name(restore_resolvers: Any) -> None:
    with raises(ValueError, match=re.escape("cannot use an empty resolver name")):
        OmegaConf.register_new_resolver("", lambda: None)


def test_register_non_inspectable_resolver(mocker: Any, restore_resolvers: Any) -> None:
    # When a function `f()` is not inspectable (e.g., some built-in CPython functions),
    # `inspect.signature(f)` will raise a `ValueError`. We want to make sure this does
    # not prevent us from using `f()` as a resolver.
    def signature_not_inspectable(*args: Any, **kw: Any) -> None:
        raise ValueError

    mocker.patch("inspect.signature", signature_not_inspectable)
    OmegaConf.register_new_resolver("not_inspectable", lambda: 123)
    assert OmegaConf.create({"x": "${not_inspectable:}"}).x == 123


@mark.parametrize(
    ("use_cache_1", "use_cache_2", "expected"),
    [
        (False, False, 2),
        (False, True, 2),
        (True, False, 2),
        (True, True, 1),  # value is obtained from cache (see also #637)
    ],
)
def test_register_resolver_with_replace(
    restore_resolvers: Any,
    use_cache_1: bool,
    use_cache_2: bool,
    expected: Any,
) -> None:
    OmegaConf.register_new_resolver("foo", lambda: 1, use_cache=use_cache_1)
    cfg = OmegaConf.create({"x": "${foo:}"})
    assert cfg.x == 1
    OmegaConf.register_new_resolver(
        "foo", lambda: 2, use_cache=use_cache_2, replace=True
    )
    assert cfg.x == expected


def test_clear_resolvers_and_has_resolver(restore_resolvers: Any) -> None:
    assert not OmegaConf.has_resolver("foo")
    OmegaConf.register_new_resolver("foo", lambda x: x + 10)
    assert OmegaConf.has_resolver("foo")
    OmegaConf.clear_resolvers()
    assert not OmegaConf.has_resolver("foo")


def test_clear_resolvers_and_has_resolver_legacy(restore_resolvers: Any) -> None:
    assert not OmegaConf.has_resolver("foo")
    OmegaConf.legacy_register_resolver("foo", lambda x: int(x) + 10)
    assert OmegaConf.has_resolver("foo")
    OmegaConf.clear_resolvers()
    assert not OmegaConf.has_resolver("foo")


def test_get_resolver_deprecation() -> None:
    with warns(
        UserWarning, match=re.escape("https://github.com/omry/omegaconf/issues/608")
    ):
        assert OmegaConf.get_resolver("foo") is None


def test_register_resolver_1(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver("plus_10", lambda x: x + 10)
    c = OmegaConf.create(
        {"k": "${plus_10:990}", "node": {"bar": 10, "foo": "${plus_10:${.bar}}"}}
    )

    assert type(c.k) == int
    assert c.k == 1000
    assert c.node.foo == 20  # this also tests relative interpolations with resolvers


def test_register_resolver_1_legacy(restore_resolvers: Any) -> None:
    OmegaConf.legacy_register_resolver("plus_10", lambda x: int(x) + 10)
    c = OmegaConf.create({"k": "${plus_10:990}"})

    assert type(c.k) == int
    assert c.k == 1000


def test_resolver_cache_1(restore_resolvers: Any) -> None:
    # The cache is important to allow embedding of functions like time()
    # without having the value change during the program execution.
    OmegaConf.register_new_resolver(
        "random", lambda _: random.randint(0, 10000000), use_cache=True
    )
    c = OmegaConf.create({"k": "${random:__}"})
    assert c.k == c.k


def test_resolver_cache_1_legacy(restore_resolvers: Any) -> None:
    OmegaConf.legacy_register_resolver("random", lambda _: random.randint(0, 10000000))
    c = OmegaConf.create({"k": "${random:_}"})
    assert c.k == c.k


def test_resolver_cache_2(restore_resolvers: Any) -> None:
    """
    Tests that resolver cache is not shared between different OmegaConf objects
    """
    OmegaConf.register_new_resolver(
        "random", lambda _: random.randint(0, 10000000), use_cache=True
    )
    c1 = OmegaConf.create({"k": "${random:__}"})
    c2 = OmegaConf.create({"k": "${random:__}"})

    assert c1.k != c2.k
    assert c1.k == c1.k
    assert c2.k == c2.k


def test_resolver_cache_2_legacy(restore_resolvers: Any) -> None:
    OmegaConf.legacy_register_resolver("random", lambda _: random.randint(0, 10000000))
    c1 = OmegaConf.create({"k": "${random:_}"})
    c2 = OmegaConf.create({"k": "${random:_}"})

    assert c1.k != c2.k
    assert c1.k == c1.k
    assert c2.k == c2.k


def test_resolver_cache_3_dict_list(restore_resolvers: Any) -> None:
    """
    Test that the resolver cache works as expected with lists and dicts.

    Note that since the cache is based on string literals, changing the order of
    items in a dictionary is considered as a different input.
    """
    OmegaConf.register_new_resolver(
        "random", lambda _: random.uniform(0, 1), use_cache=True
    )
    c = OmegaConf.create(
        {
            "lst1": "${random:[0, 1]}",
            "lst2": "${random:[0, 1]}",
            "lst3": "${random:[]}",
            "dct1": "${random:{a: 1, b: 2}}",
            "dct2": "${random:{b: 2, a: 1}}",
            "mixed1": "${random:{x: [1.1], y: {a: true, b: false, c: null, d: []}}}",
            "mixed2": "${random:{x: [1.1], y: {b: false, c: null, a: true, d: []}}}",
        }
    )
    assert c.lst1 == c.lst1
    assert c.lst1 == c.lst2
    assert c.lst1 != c.lst3
    assert c.dct1 == c.dct1
    assert c.dct1 != c.dct2
    assert c.mixed1 == c.mixed1
    assert c.mixed2 == c.mixed2
    assert c.mixed1 != c.mixed2


def test_resolver_cache_4_interpolation(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver("test", lambda x: x, use_cache=True)
    c = OmegaConf.create({"x": "${test:${y}}", "y": 0})

    assert c.x == 0
    c.y = 1
    assert c.x == 0  # cache is based on string literals


def test_resolver_no_cache(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver(
        "random", lambda _: random.uniform(0, 1), use_cache=False
    )
    c = OmegaConf.create({"k": "${random:__}"})
    assert c.k != c.k


def test_resolver_dot_start(common_resolvers: Any) -> None:
    """
    Regression test for #373
    """
    c = OmegaConf.create(
        {"foo_nodot": "${identity:bar}", "foo_dot": "${identity:.bar}"}
    )
    assert c.foo_nodot == "bar"
    assert c.foo_dot == ".bar"


def test_resolver_dot_start_legacy(common_resolvers: Any) -> None:
    c = OmegaConf.create(
        {"foo_nodot": "${identity:bar}", "foo_dot": "${identity:.bar}"}
    )
    assert c.foo_nodot == "bar"
    assert c.foo_dot == ".bar"


@mark.parametrize(
    "resolver,name,key,result",
    [
        (lambda *args: args, "arg_list", "${my_resolver:cat, dog}", ("cat", "dog")),
        (
            lambda *args: args,
            "escape_comma",
            "${my_resolver:cat\\, do g}",
            ("cat, do g",),
        ),
        (
            lambda *args: args,
            "escape_whitespace",
            "${my_resolver:cat,\\ do g}",
            ("cat", " do g"),
        ),
        (lambda: "zero", "zero_arg", "${my_resolver:}", "zero"),
    ],
)
def test_resolver_that_allows_a_list_of_arguments(
    restore_resolvers: Any, resolver: Resolver, name: str, key: str, result: Any
) -> None:
    OmegaConf.register_new_resolver("my_resolver", resolver)
    c = OmegaConf.create({name: key})
    assert c[name] == result


@mark.parametrize(
    "resolver,name,key,result",
    [
        (lambda *args: args, "arg_list", "${my_resolver:cat, dog}", ("cat", "dog")),
        (
            lambda *args: args,
            "escape_comma",
            "${my_resolver:cat\\, do g}",
            ("cat, do g",),
        ),
        (
            lambda *args: args,
            "escape_whitespace",
            "${my_resolver:cat,\\ do g}",
            ("cat", " do g"),
        ),
        (lambda: "zero", "zero_arg", "${my_resolver:}", "zero"),
    ],
)
def test_resolver_that_allows_a_list_of_arguments_legacy(
    restore_resolvers: Any, resolver: Resolver, name: str, key: str, result: Any
) -> None:
    OmegaConf.legacy_register_resolver("my_resolver", resolver)
    c = OmegaConf.create({name: key})
    assert c[name] == result


def test_resolver_deprecated_behavior(restore_resolvers: Any) -> None:
    # Ensure that resolvers registered with the old "register_resolver()" function
    # behave as expected.

    # The registration should trigger a deprecation warning.
    with warns(UserWarning):
        OmegaConf.register_resolver("my_resolver", lambda *args: args)

    c = OmegaConf.create(
        {
            "int": "${my_resolver:1}",
            "null": "${my_resolver:null}",
            "bool": "${my_resolver:TruE,falSE}",
            "str": "${my_resolver:a,b,c}",
            "inter": "${my_resolver:${int}}",
        }
    )

    # All resolver arguments should be provided as strings (with no modification).
    assert c.int == ("1",)
    assert c.null == ("null",)
    assert c.bool == ("TruE", "falSE")
    assert c.str == ("a", "b", "c")

    # Trying to nest interpolations should trigger an error (users should switch to
    # `register_new_resolver()` in order to use nested interpolations).
    with raises(ValueError):
        c.inter


def test_copy_cache(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver(
        "random", lambda _: random.randint(0, 10000000), use_cache=True
    )
    d = {"k": "${random:__}"}
    c1 = OmegaConf.create(d)
    assert c1.k == c1.k

    c2 = OmegaConf.create(d)
    assert c2.k != c1.k
    OmegaConf.set_cache(c2, OmegaConf.get_cache(c1))
    assert c2.k == c1.k

    c3 = OmegaConf.create(d)

    assert c3.k != c1.k
    OmegaConf.copy_cache(c1, c3)
    assert c3.k == c1.k


def test_clear_cache(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver("random", lambda _: random.randint(0, 10000000))
    c = OmegaConf.create({"k": "${random:__}"})
    old = c.k
    OmegaConf.clear_cache(c)
    assert old != c.k


@mark.parametrize("readonly", [True, False])
def test_resolver_output_dict(restore_resolvers: Any, readonly: bool) -> None:
    some_dict = {"a": 0, "b": "${y}"}
    OmegaConf.register_new_resolver("dict", lambda: some_dict)
    c = OmegaConf.create({"x": "${dict:}", "y": -1})
    OmegaConf.set_readonly(c, readonly)
    assert isinstance(c.x, dict)
    assert c.x == some_dict
    x_node = dereference_node(c, "x")
    assert isinstance(x_node, InterpolationResultNode)
    assert x_node._get_flag("readonly")


@mark.parametrize("readonly", [True, False])
@mark.parametrize(
    ("data", "expected_type"),
    [
        param({"a": 0, "b": "${y}"}, dict, id="dict"),
        param(["a", 0, "${y}"], list, id="list"),
    ],
)
def test_resolver_output_plain_dict_list(
    restore_resolvers: Any, readonly: bool, data: Any, expected_type: type
) -> None:
    OmegaConf.register_new_resolver("get_data", lambda: data)
    c = OmegaConf.create({"x": "${get_data:}", "y": -1})
    OmegaConf.set_readonly(c, readonly)

    assert isinstance(c.x, expected_type)
    assert c.x == data

    x_node = dereference_node(c, "x")
    assert isinstance(x_node, InterpolationResultNode)
    assert x_node._get_flag("readonly")


def test_register_cached_resolver_with_keyword_unsupported() -> None:
    with raises(ValueError):
        OmegaConf.register_new_resolver("root", lambda _root_: None, use_cache=True)
    with raises(ValueError):
        OmegaConf.register_new_resolver("parent", lambda _parent_: None, use_cache=True)


def test_resolver_with_parent(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver(
        "parent", lambda _parent_: _parent_, use_cache=False
    )

    cfg = OmegaConf.create(
        {
            "a": 10,
            "b": {
                "c": 20,
                "parent": "${parent:}",
            },
            "parent": "${parent:}",
        }
    )

    assert cfg.parent is cfg
    assert cfg.b.parent is cfg.b


def test_resolver_with_root(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver("root", lambda _root_: _root_, use_cache=False)
    cfg = OmegaConf.create(
        {
            "a": 10,
            "b": {
                "c": 20,
                "root": "${root:}",
            },
            "root": "${root:}",
        }
    )

    assert cfg.root is cfg
    assert cfg.b.root is cfg


def test_resolver_with_root_and_parent(restore_resolvers: Any) -> None:
    OmegaConf.register_new_resolver(
        "both", lambda _root_, _parent_: _root_.add + _parent_.add, use_cache=False
    )

    cfg = OmegaConf.create(
        {
            "add": 10,
            "b": {
                "add": 20,
                "both": "${both:}",
            },
            "both": "${both:}",
        }
    )
    assert cfg.both == 20
    assert cfg.b.both == 30


def test_resolver_with_parent_and_default_value(restore_resolvers: Any) -> None:
    def parent_and_default(default: int = 10, *, _parent_: Any) -> Any:
        return _parent_.add + default

    OmegaConf.register_new_resolver(
        "parent_and_default", parent_and_default, use_cache=False
    )

    cfg = OmegaConf.create(
        {
            "add": 10,
            "no_param": "${parent_and_default:}",
            "param": "${parent_and_default:20}",
        }
    )

    assert cfg.no_param == 20
    assert cfg.param == 30


@mark.parametrize(
    ("cfg2", "expected"),
    [
        param({"foo": {"b": 1}}, {"foo": {"a": 0, "b": 1}}, id="extend"),
        param({"foo": {"b": "${.a}"}}, {"foo": {"a": 0, "b": 0}}, id="extend_inter"),
        param({"foo": {"a": 1}}, {"foo": {"a": 1}}, id="override_int"),
        param({"foo": {"a": {"b": 1}}}, {"foo": {"a": {"b": 1}}}, id="override_dict"),
        param({"foo": 10}, {"foo": 10}, id="replace_interpolation"),
        param({"bar": 10}, {"foo": {"a": 0}, "bar": 10}, id="other_node"),
    ],
)
def test_merge_into_resolver_output(
    restore_resolvers: Any, cfg2: Any, expected: Any
) -> None:
    OmegaConf.register_new_resolver(
        "make", lambda _parent_: OmegaConf.create({"a": 0}, parent=_parent_)
    )

    cfg = OmegaConf.create({"foo": "${make:}"})
    assert OmegaConf.merge(cfg, cfg2) == expected



import builtins
from typing import Any

from pytest import fixture, mark, param

from omegaconf import (
    AnyNode,
    BooleanNode,
    Container,
    DictConfig,
    EnumNode,
    FloatNode,
    IntegerNode,
    ListConfig,
    Node,
    OmegaConf,
    StringNode,
    ValueNode,
)
from omegaconf._utils import type_str
from pydevd_plugins.extensions.pydevd_plugin_omegaconf import (
    OmegaConfDeveloperResolver,
    OmegaConfUserResolver,
)
from tests import Color


@fixture
def resolver() -> Any:
    yield OmegaConfUserResolver()


@mark.parametrize(
    ("obj", "expected"),
    [
        # nodes
        param(AnyNode(10), {}, id="any:10"),
        param(StringNode("foo"), {}, id="str:foo"),
        param(IntegerNode(10), {}, id="int:10"),
        param(FloatNode(3.14), {}, id="float:3.14"),
        param(BooleanNode(True), {}, id="bool:True"),
        param(EnumNode(enum_type=Color, value=Color.RED), {}, id="Color:Color.RED"),
        # nodes are never returning a dictionary
        param(AnyNode("${foo}", parent=DictConfig({"foo": 10})), {}, id="any:inter_10"),
        # DictConfig
        param(DictConfig({"a": 10}), {"a": AnyNode(10)}, id="dict"),
        param(
            DictConfig({"a": 10, "b": "${a}"}),
            {"a": AnyNode(10), "b": AnyNode("${a}")},
            id="dict:interpolation_value",
        ),
        param(
            DictConfig({"a": 10, "b": "${zzz}"}),
            {"a": AnyNode(10), "b": AnyNode("${zzz}")},
            id="dict:interpolation_value_error",
        ),
        param(
            DictConfig({"a": 10, "b": "foo_${a}"}),
            {"a": AnyNode(10), "b": AnyNode("foo_${a}")},
            id="dict:str_interpolation_value",
        ),
        param(DictConfig("${zzz}"), {}, id="dict:inter_error"),
        # ListConfig
        param(
            ListConfig(["a", "b"]), {"0": AnyNode("a"), "1": AnyNode("b")}, id="list"
        ),
        param(
            ListConfig(["${1}", 10]),
            {"0": AnyNode("${1}"), "1": AnyNode(10)},
            id="list:interpolation_value",
        ),
        param(ListConfig("${zzz}"), {}, id="list:inter_error"),
    ],
)
def test_get_dictionary_node(resolver: Any, obj: Any, expected: Any) -> None:
    res = resolver.get_dictionary(obj)
    assert res == expected


@mark.parametrize(
    ("obj", "attribute", "expected"),
    [
        # dictconfig
        param(DictConfig({"a": 10}), "a", AnyNode(10), id="dict"),
        param(
            DictConfig({"a": DictConfig(None)}),
            "a",
            DictConfig(None),
            id="dict:none",
        ),
        # listconfig
        param(ListConfig([10]), 0, AnyNode(10), id="list"),
        param(ListConfig(["???"]), 0, AnyNode("???"), id="list:missing_item"),
    ],
)
def test_resolve(
    resolver: Any,
    obj: Any,
    attribute: str,
    expected: Any,
) -> None:
    res = resolver.resolve(obj, attribute)
    assert res == expected
    assert type(res) is type(expected)


@mark.parametrize(
    ("obj", "attribute", "expected"),
    [
        param(
            OmegaConf.create({"a": 10, "inter": "${a}"}), "inter", {}, id="dict:inter"
        ),
        param(
            OmegaConf.create({"missing": "???"}),
            "missing",
            {},
            id="dict:missing_value",
        ),
        param(
            OmegaConf.create({"none": None}),
            "none",
            {},
            id="dict:none_value",
        ),
        param(
            OmegaConf.create({"none": DictConfig(None)}),
            "none",
            {},
            id="dict:none_dictconfig_value",
        ),
        param(
            OmegaConf.create({"missing": DictConfig("???")}),
            "missing",
            {},
            id="dict:missing_dictconfig_value",
        ),
        param(
            OmegaConf.create({"a": {"b": 10}, "b": DictConfig("${a}")}),
            "b",
            {"b": 10},
            id="dict:interpolation_dictconfig_value",
        ),
    ],
)
def test_get_dictionary_dictconfig(
    resolver: Any,
    obj: Any,
    attribute: str,
    expected: Any,
) -> None:
    field = resolver.resolve(obj, attribute)
    res = resolver.get_dictionary(field)
    assert res == expected
    assert type(res) is type(expected)


@mark.parametrize(
    ("obj", "attribute", "expected"),
    [
        param(
            ListConfig("${list}", parent=DictConfig({"list": [{"a": 10}]})),
            "0",
            {"a": 10},
            id="inter_list:dict_element",
        ),
        param(
            DictConfig("${dict}", parent=DictConfig({"dict": {"a": {"b": 10}}})),
            "a",
            {"b": 10},
            id="inter_dict:dict_element",
        ),
    ],
)
def test_resolve_through_container_interpolation(
    resolver: Any, obj: Any, attribute: str, expected: Any
) -> None:
    res = resolver.resolve(obj, attribute)
    assert res == expected


@mark.parametrize(
    ("obj", "attribute", "expected"),
    [
        param(OmegaConf.create(["${.1}", 10]), "0", {}, id="list:inter_value"),
        param(
            OmegaConf.create({"a": ListConfig(None)}),
            "a",
            {},
            id="list:none_listconfig_value",
        ),
        param(
            OmegaConf.create({"a": ListConfig("???")}),
            "a",
            {},
            id="list:missing_listconfig_value",
        ),
        param(
            OmegaConf.create({"a": [1, 2], "b": ListConfig("${a}")}),
            "b",
            {"0": 1, "1": 2},
            id="list:interpolation_listconfig_value",
        ),
    ],
)
def test_get_dictionary_listconfig(
    resolver: Any,
    obj: Any,
    attribute: str,
    expected: Any,
) -> None:
    field = resolver.resolve(obj, attribute)
    res = resolver.get_dictionary(field)
    assert res == expected
    assert type(res) is type(expected)


@mark.parametrize("resolver", [OmegaConfUserResolver(), OmegaConfDeveloperResolver()])
@mark.parametrize(
    ("type_", "expected"),
    [
        # containers
        (Container, True),
        (DictConfig, True),
        (ListConfig, True),
        # nodes
        (Node, True),
        (ValueNode, True),
        (AnyNode, True),
        (IntegerNode, True),
        (FloatNode, True),
        (StringNode, True),
        (BooleanNode, True),
        (EnumNode, True),
        # not covering some other things.
        (builtins.int, False),
        (dict, False),
        (list, False),
    ],
)
def test_can_provide(resolver: Any, type_: Any, expected: bool) -> None:
    assert resolver.can_provide(type_, type_str(type_)) == expected


@mark.parametrize(
    ("obj", "expected"),
    [
        (AnyNode(10), "10"),
        (AnyNode("???"), "??? <MISSING>"),
        (
            AnyNode("${foo}", parent=OmegaConf.create({})),
            "${foo} -> ERR: Interpolation key 'foo' not found",
        ),
        (AnyNode("${foo}", parent=OmegaConf.create({"foo": 10})), "${foo} -> 10"),
        (
            DictConfig("${foo}", parent=OmegaConf.create({"foo": {"a": 10}})),
            "${foo} -> {'a': 10}",
        ),
        (
            ListConfig("${foo}", parent=OmegaConf.create({"foo": [1, 2]})),
            "${foo} -> [1, 2]",
        ),
    ],
)
def test_get_str(resolver: Any, obj: Any, expected: str) -> None:
    assert resolver.get_str(obj) == expected


def test_dev_resolver() -> None:
    resolver = OmegaConfDeveloperResolver()
    cfg = OmegaConf.create({"foo": 10})
    assert resolver.resolve(cfg, "_metadata") is cfg.__dict__["_metadata"]
    assert resolver.get_dictionary(cfg) is cfg.__dict__



import os
import warnings
from typing import Any, Optional

from omegaconf._utils import decode_primitive
from omegaconf.errors import ValidationError
from omegaconf.resolvers import oc


# DEPRECATED: remove in 2.2
def env(key: str, default: Optional[str] = None) -> Any:
    warnings.warn(
        "The `env` resolver is deprecated, see https://github.com/omry/omegaconf/issues/573"
    )

    try:
        return decode_primitive(os.environ[key])
    except KeyError:
        if default is not None:
            return decode_primitive(default)
        else:
            raise ValidationError(f"Environment variable '{key}' not found")


__all__ = [
    "env",
    "oc",
]



import re
from typing import Any, List

from pytest import mark, param, raises

from omegaconf import DictConfig, ListConfig, OmegaConf
from omegaconf.errors import (
    InterpolationResolutionError,
    InterpolationToMissingValueError,
)
from omegaconf.resolvers.oc.dict import _get_and_validate_dict_input
from tests import User, Users


@mark.parametrize(
    ("cfg", "key", "expected"),
    [
        param(
            {"foo": "${oc.dict.keys:bar}", "bar": {"a": 0, "b": 1}},
            "foo",
            ["a", "b"],
            id="dictconfig",
        ),
        param(
            {"foo": "${oc.dict.keys:bar}", "bar": "${boz}", "boz": {"a": 0, "b": 1}},
            "foo",
            ["a", "b"],
            id="dictconfig_chained_interpolation",
        ),
        param(
            {"a": "${oc.dict.keys:''}", "b": 10},
            "a",
            ["a", "b"],
            id="select_keys_of_root",
        ),
    ],
)
def test_dict_keys(cfg: Any, key: Any, expected: Any) -> None:
    cfg = OmegaConf.create(cfg)
    val = cfg[key]
    assert val == expected
    assert isinstance(val, ListConfig)
    assert val._parent is cfg


@mark.parametrize(
    ("cfg", "key", "expected"),
    [
        param(
            {"x": "${oc.dict.keys_or_values:y}", "y": "???"},
            "x",
            raises(
                InterpolationResolutionError,
                match=re.escape(
                    "MissingMandatoryValue raised while resolving interpolation: "
                    "Missing mandatory value: y"
                ),
            ),
            id="select_missing",
        ),
        param(
            {"foo": "${oc.dict.keys_or_values:bar}"},
            "foo",
            raises(
                InterpolationResolutionError,
                match=re.escape(
                    "ConfigKeyError raised while resolving interpolation: "
                    "Key not found: 'bar'"
                ),
            ),
            id="config_key_error",
        ),
        param(
            {"foo": "${oc.dict.keys_or_values:bar}", "bar": 0},
            "foo",
            raises(
                InterpolationResolutionError,
                match=re.escape(
                    "TypeError raised while resolving interpolation: "
                    "`oc.dict.keys_or_values` cannot be applied to objects of type: int"
                ),
            ),
            id="type_error",
        ),
        param(
            {"foo": "${oc.dict.keys_or_values:bar}", "bar": DictConfig(None)},
            "foo",
            raises(
                InterpolationResolutionError,
                match=re.escape(
                    "TypeError raised while resolving interpolation: "
                    "`oc.dict.keys_or_values` cannot be applied to objects of type: NoneType"
                ),
            ),
            id="type_error_dictconfig",
        ),
    ],
)
def test_get_and_validate_dict_input(
    restore_resolvers: Any, cfg: Any, key: Any, expected: Any
) -> None:
    OmegaConf.register_new_resolver(
        "oc.dict.keys_or_values",
        lambda in_dict, _parent_: _get_and_validate_dict_input(
            in_dict, parent=_parent_, resolver_name="oc.dict.keys_or_values"
        ),
    )
    cfg = OmegaConf.create(cfg)
    with expected:
        cfg[key]


@mark.parametrize(
    ("cfg", "key", "expected_val", "expected_content"),
    [
        param(
            {"foo": "${oc.dict.values:bar}", "bar": {"a": 0, "b": 1}},
            "foo",
            [0, 1],
            ["${bar.a}", "${bar.b}"],
            id="dictconfig",
        ),
        param(
            {
                "foo": "${oc.dict.values:bar}",
                "bar": {"a": {"x": 0, "y": 1}, "b": {"x": 0}},
            },
            "foo",
            [{"x": 0, "y": 1}, {"x": 0}],
            ["${bar.a}", "${bar.b}"],
            id="dictconfig_deep",
        ),
        param(
            {
                "foo": "${oc.dict.values:bar}",
                "bar": {"key": "${val_ref}"},
                "val_ref": "value",
            },
            "foo",
            ["value"],
            ["${bar.key}"],
            id="dictconfig_with_interpolated_value",
        ),
        param(
            {
                "foo": "${oc.dict.values:bar}",
                "bar": "${boz}",
                "boz": {"a": 0, "b": 1},
            },
            "foo",
            [0, 1],
            ["${bar.a}", "${bar.b}"],
            id="dictconfig_chained_interpolation",
        ),
    ],
)
def test_dict_values(
    cfg: Any, key: Any, expected_val: Any, expected_content: Any
) -> None:

    cfg = OmegaConf.create(cfg)
    val = cfg[key]
    assert val == expected_val
    assert isinstance(val, ListConfig)
    assert val._parent is cfg
    content = val._content
    assert content == expected_content


def test_dict_values_with_missing_value() -> None:
    cfg = OmegaConf.create({"foo": "${oc.dict.values:bar}", "bar": {"missing": "???"}})
    foo = cfg.foo
    with raises(InterpolationToMissingValueError):
        foo[0]
    cfg.bar.missing = 1
    assert foo[0] == 1


@mark.parametrize(
    ("make_resolver", "key", "expected"),
    [
        param(
            lambda _parent_: OmegaConf.create({"x": 1}, parent=_parent_),
            0,
            1,
            id="basic",
        ),
        param(
            lambda _parent_: OmegaConf.create({"x": "${y}", "y": 1}, parent=_parent_),
            0,
            999,  # referring to the `y` node from the global config
            id="inter_abs",
        ),
        param(
            lambda _parent_: OmegaConf.create({"x": "${.y}", "y": 1}, parent=_parent_),
            0,
            1,
            id="inter_rel",
        ),
        param(lambda: OmegaConf.create({"x": 1}), 0, 1, id="basic_no_parent"),
        param(
            lambda: OmegaConf.create({"x": "${y}", "y": 1}),
            0,
            1,  # no parent => referring to the local `y` node in the generated config
            id="inter_abs_no_parent",
        ),
        param(
            lambda: OmegaConf.create({"x": "${.y}", "y": 1}),
            0,
            1,
            id="inter_rel_no_parent",
        ),
    ],
)
def test_dict_values_dictconfig_resolver_output(
    restore_resolvers: Any, make_resolver: Any, key: Any, expected: Any
) -> None:
    OmegaConf.register_new_resolver("make", make_resolver)
    cfg = OmegaConf.create(
        {
            "foo": "${oc.dict.values:bar}",
            "bar": "${make:}",
            "y": 999,
        }
    )

    assert cfg.foo[key] == expected


def test_dict_values_are_typed() -> None:
    cfg = OmegaConf.create(
        {
            "x": "${oc.dict.values: y.name2user}",
            "y": Users(
                name2user={
                    "john": User(name="john", age=30),
                    "jane": User(name="jane", age=33),
                }
            ),
        }
    )
    x = cfg.x
    assert x._metadata.ref_type == List[User]
    assert x._metadata.element_type == User


@mark.parametrize(
    ("cfg", "expected"),
    [
        param({"x": "${oc.dict.values:y}", "y": {"a": 1}}, [1], id="values_inter"),
        param({"x": "${oc.dict.keys:y}", "y": {"a": 1}}, ["a"], id="keys_inter"),
    ],
)
def test_readonly_parent(cfg: Any, expected: Any) -> None:
    cfg = OmegaConf.create(cfg)
    OmegaConf.set_readonly(cfg, True)
    assert cfg.x == expected


@mark.parametrize(
    ("cfg", "expected"),
    [
        param(
            {"x": "${sum:${oc.dict.values:y}}", "y": {"one": 1, "two": 2}},
            3,
            id="values",
        ),
        param(
            {"x": "${sum:${oc.dict.keys:y}}", "y": {1: "one", 2: "two"}},
            3,
            id="keys",
        ),
    ],
)
def test_nested_oc_dict(restore_resolvers: Any, cfg: Any, expected: Any) -> None:
    OmegaConf.register_new_resolver("sum", sum)
    cfg = OmegaConf.create(cfg)
    assert cfg.x == expected


@mark.parametrize(
    "cfg",
    [
        param({"x": "${oc.dict.keys:[]}"}, id="list"),
        param({"x": "${oc.dict.keys:${bool}}", "bool": True}, id="bool_interpolation"),
        param({"x": "${oc.dict.keys:int}", "int": 0}, id="int_select"),
    ],
)
def test_dict_keys_invalid_type(cfg: Any) -> None:
    cfg = OmegaConf.create(cfg)
    with raises(InterpolationResolutionError, match="TypeError"):
        cfg.x


@mark.parametrize(
    "cfg",
    [
        param({"x": "${oc.dict.values:[]}"}, id="list"),
        param(
            {"x": "${oc.dict.values:${bool}}", "bool": True}, id="bool_interpolation"
        ),
        param({"x": "${oc.dict.values:int}", "int": 0}, id="int_select"),
    ],
)
def test_dict_values_invalid_type(cfg: Any) -> None:
    cfg = OmegaConf.create(cfg)
    with raises(InterpolationResolutionError, match="TypeError"):
        cfg.x



import re
from importlib import import_module
from typing import Any, Optional

from pytest import fixture, mark, raises

from omegaconf import (
    DictConfig,
    IntegerNode,
    Node,
    OmegaConf,
    ValidationError,
    _utils,
    flag_override,
)
from omegaconf._utils import _is_optional, get_ref_type
from omegaconf.errors import ConfigKeyError, UnsupportedValueType
from tests import IllegalType


@fixture(
    params=[
        "tests.structured_conf.data.dataclasses",
        "tests.structured_conf.data.attr_classes",
    ],
    ids=["dataclasses", "attr_classes"],
)
def module(request: Any) -> Any:
    return import_module(request.param)


class TestStructured:
    class TestBasic:
        def test_error_on_non_structured_config_class(self, module: Any) -> None:
            with raises(ValidationError, match="structured config"):
                OmegaConf.structured(module.NotStructuredConfig)

        def test_error_on_non_structured_nested_config_class(self, module: Any) -> None:
            with raises(
                ValidationError,
                match=re.escape("Unexpected object type: NotStructuredConfig"),
            ):
                OmegaConf.structured(module.StructuredWithInvalidField)

            ret = OmegaConf.structured(
                module.StructuredWithInvalidField, flags={"allow_objects": True}
            )
            assert list(ret.keys()) == ["bar"]
            assert ret.bar == module.NotStructuredConfig()

        def test_error_on_creation_with_bad_value_type(self, module: Any) -> None:
            with raises(
                ValidationError,
                match=re.escape("Value 'seven' could not be converted to Integer"),
            ):
                OmegaConf.structured(module.User(age="seven"))

        def test_assignment_of_subclass(self, module: Any) -> None:
            cfg = OmegaConf.create({"plugin": module.Plugin})
            cfg.plugin = OmegaConf.structured(module.ConcretePlugin)
            assert OmegaConf.get_type(cfg.plugin) == module.ConcretePlugin
            assert (
                OmegaConf.get_type(cfg.plugin.params)
                == module.ConcretePlugin.FoobarParams
            )

        def test_assignment_of_non_subclass_1(self, module: Any) -> None:
            cfg = OmegaConf.create(
                {"plugin": DictConfig(module.Plugin, ref_type=module.Plugin)}
            )
            with raises(ValidationError):
                cfg.plugin = OmegaConf.structured(module.FaultyPlugin)

        def test_merge(self, module: Any) -> None:
            cfg1 = OmegaConf.create({"plugin": module.Plugin})
            cfg2 = OmegaConf.create({"plugin": module.ConcretePlugin})
            assert cfg2.plugin == module.ConcretePlugin
            res: Any = OmegaConf.merge(cfg1, cfg2)
            assert OmegaConf.get_type(res.plugin) == module.ConcretePlugin
            assert (
                OmegaConf.get_type(res.plugin.params)
                == module.ConcretePlugin.FoobarParams
            )

        def test_merge_of_non_subclass_1(self, module: Any) -> None:
            cfg1 = OmegaConf.create({"plugin": module.Plugin})
            cfg2 = OmegaConf.create({"plugin": module.FaultyPlugin})
            with raises(ValidationError):
                OmegaConf.merge(cfg1, cfg2)

        def test_merge_error_new_attribute(self, module: Any) -> None:
            cfg = OmegaConf.structured(module.ConcretePlugin)
            cfg2 = OmegaConf.create({"params": {"bar": 10}})
            # raise if an invalid key is merged into a struct
            with raises(ConfigKeyError):
                OmegaConf.merge(cfg, cfg2)

        def test_merge_error_override_bad_type(self, module: Any) -> None:
            cfg = OmegaConf.structured(module.ConcretePlugin)

            # raise if an invalid key is merged into a struct
            with raises(ValidationError):
                OmegaConf.merge(cfg, {"params": {"foo": "zonk"}})

        def test_error_message(self, module: Any) -> None:
            cfg = OmegaConf.structured(module.StructuredOptional)
            msg = re.escape("child 'not_optional' is not Optional")
            with raises(ValidationError, match=msg):
                cfg.not_optional = None

    def test_none_assignment(self, module: Any) -> None:
        cfg = OmegaConf.create({"plugin": module.Plugin})
        # can assign None to params (type Any):
        cfg.plugin.params = None
        assert cfg.plugin.params is None

        cfg2 = OmegaConf.create({"plugin": module.ConcretePlugin})
        with raises(ValidationError):
            cfg2.plugin.params = None

    @mark.parametrize("rhs", [1, "foo"])
    class TestFailedAssignmentOrMerges:
        def test_assignment_of_non_subclass_2(self, module: Any, rhs: Any) -> None:
            cfg = OmegaConf.create(
                {"plugin": DictConfig(module.Plugin, ref_type=module.Plugin)}
            )
            with raises(ValidationError):
                cfg.plugin = rhs

        def test_merge_of_non_subclass_2(self, module: Any, rhs: Any) -> None:
            cfg1 = OmegaConf.create({"plugin": module.Plugin})
            cfg2 = OmegaConf.create({"plugin": rhs})
            with raises(ValidationError):
                OmegaConf.merge(cfg1, cfg2)

    def test_get_type(self, module: Any) -> None:
        cfg1 = OmegaConf.create(module.LinkedList)
        assert OmegaConf.get_type(cfg1) == module.LinkedList
        assert _utils.get_ref_type(cfg1, "next") == Optional[module.LinkedList]
        assert OmegaConf.get_type(cfg1, "next") is None

        assert cfg1.next is None
        assert OmegaConf.is_missing(cfg1, "value")

        cfg2 = OmegaConf.create(module.MissingTest.Missing1)
        assert OmegaConf.is_missing(cfg2, "head")
        assert _utils.get_ref_type(cfg2, "head") == module.LinkedList
        assert OmegaConf.get_type(cfg2, "head") is None

    def test_merge_structured_onto_dict(self, module: Any) -> None:
        c1 = OmegaConf.create({"name": 7})
        c2 = OmegaConf.merge(c1, module.User)
        assert c1 == {"name": 7}
        # type of name becomes str
        assert c2 == {"name": "7", "age": "???"}

    def test_merge_structured_onto_dict_nested(self, module: Any) -> None:
        c1 = OmegaConf.create({"user": {"name": 7}})
        c2 = OmegaConf.merge(c1, module.MissingUserField)
        assert c1 == {"user": {"name": 7}}
        # type of name becomes str
        assert c2 == {"user": {"name": "7", "age": "???"}}
        assert isinstance(c2, DictConfig)
        assert get_ref_type(c2, "user") == module.User

    def test_merge_structured_onto_dict_nested2(self, module: Any) -> None:
        c1 = OmegaConf.create({"user": {"name": IntegerNode(value=7)}})
        c2 = OmegaConf.merge(c1, module.MissingUserField)
        assert c1 == {"user": {"name": 7}}
        # type of name remains int
        assert c2 == {"user": {"name": 7, "age": "???"}}
        assert isinstance(c2, DictConfig)
        assert get_ref_type(c2, "user") == module.User

    def test_merge_structured_onto_dict_nested3(self, module: Any) -> None:
        c1 = OmegaConf.create({"user": {"name": "alice"}})
        c2 = OmegaConf.merge(c1, module.MissingUserWithDefaultNameField)
        assert c1 == {"user": {"name": "alice"}}
        # name is not changed
        assert c2 == {"user": {"name": "alice", "age": "???"}}
        assert isinstance(c2, DictConfig)
        assert get_ref_type(c2, "user") == module.UserWithDefaultName

    def test_merge_missing_object_onto_typed_dictconfig(self, module: Any) -> None:
        c1 = OmegaConf.structured(module.DictOfObjects)
        c2 = OmegaConf.merge(c1, {"users": {"bob": "???"}})
        assert isinstance(c2, DictConfig)
        assert OmegaConf.is_missing(c2.users, "bob")

    def test_merge_into_missing_sc(self, module: Any) -> None:
        c1 = OmegaConf.structured(module.PluginHolder)
        c2 = OmegaConf.merge(c1, {"plugin": "???"})
        assert c2.plugin == module.Plugin()

    def test_merge_missing_key_onto_structured_none(self, module: Any) -> None:
        c1 = OmegaConf.create({"foo": OmegaConf.structured(module.OptionalUser)})
        src = OmegaConf.create({"foo": {"user": "???"}})
        c2 = OmegaConf.merge(c1, src)
        assert c1.foo.user is None
        assert c2.foo.user is None
        assert c2.foo._get_node("user")._metadata.ref_type == module.User

    def test_merge_optional_structured_onto_dict(self, module: Any) -> None:
        c1 = OmegaConf.create({"user": {"name": "bob"}})
        c2 = OmegaConf.merge(c1, module.OptionalUser(module.User(name="alice")))
        assert c2.user.name == "alice"
        assert get_ref_type(c2, "user") == Optional[module.User]
        assert isinstance(c2, DictConfig)
        c2_user = c2._get_node("user")
        assert isinstance(c2_user, Node)
        # Compared to the previous assert, here we verify that the `ref_type` found
        # in the metadata is *not* optional: instead, the `optional` flag must be set.
        assert c2_user._metadata.ref_type == module.User
        assert c2_user._metadata.optional

    def test_merge_structured_interpolation_onto_dict(self, module: Any) -> None:
        c1 = OmegaConf.create(
            {
                "user_1": {"name": "bob"},
                "user_2": {"name": "alice"},
                "user_3": {"name": "joe"},
            }
        )
        src = OmegaConf.create({"user_2": module.User(), "user_3": module.User()})
        src.user_2 = "${user_1}"
        src.user_3 = None
        c2 = OmegaConf.merge(c1, src)
        assert c2.user_2.name == "bob"
        assert get_ref_type(c2, "user_2") == Any
        assert c2.user_3 is None
        assert get_ref_type(c2, "user_3") == Any

    class TestMissing:
        def test_missing1(self, module: Any) -> None:
            cfg = OmegaConf.create(module.MissingTest.Missing1)
            assert OmegaConf.is_missing(cfg, "head")

            assert OmegaConf.get_type(cfg, "head") is None

            with raises(ValidationError):
                cfg.head = 10

        def test_missing2(self, module: Any) -> None:
            cfg = OmegaConf.create(module.MissingTest.Missing2)
            assert cfg == {"head": {"next": "???", "value": 1}}
            assert OmegaConf.is_missing(cfg.head, "next")

            cfg.head.next = module.LinkedList(value=2)
            assert cfg == {"head": {"next": {"next": None, "value": 2}, "value": 1}}

        def test_plugin_holder(self, module: Any) -> None:
            cfg = OmegaConf.create(module.PluginHolder)

            assert _is_optional(cfg, "none")
            assert _utils.get_ref_type(cfg, "none") == Optional[module.Plugin]
            assert OmegaConf.get_type(cfg, "none") is None

            assert not _is_optional(cfg, "missing")
            assert _utils.get_ref_type(cfg, "missing") == module.Plugin
            assert OmegaConf.get_type(cfg, "missing") is None

            assert not _is_optional(cfg, "plugin")
            assert _utils.get_ref_type(cfg, "plugin") == module.Plugin
            assert OmegaConf.get_type(cfg, "plugin") == module.Plugin

            cfg.plugin = module.ConcretePlugin()
            assert not _is_optional(cfg, "plugin")
            assert _utils.get_ref_type(cfg, "plugin") == module.Plugin
            assert OmegaConf.get_type(cfg, "plugin") == module.ConcretePlugin

            assert not _is_optional(cfg, "plugin2")
            assert _utils.get_ref_type(cfg, "plugin2") == module.Plugin
            assert OmegaConf.get_type(cfg, "plugin2") == module.ConcretePlugin

        def test_plugin_merge(self, module: Any) -> None:
            plugin = OmegaConf.structured(module.Plugin)
            concrete = OmegaConf.structured(module.ConcretePlugin)
            ret = OmegaConf.merge(plugin, concrete)
            assert ret == concrete
            assert OmegaConf.get_type(ret) == module.ConcretePlugin

        def test_plugin_merge_2(self, module: Any) -> None:
            plugin = OmegaConf.structured(module.Plugin)
            more_fields = OmegaConf.structured(module.PluginWithAdditionalField)
            ret = OmegaConf.merge(plugin, more_fields)
            assert ret == more_fields
            assert OmegaConf.get_type(ret) == module.PluginWithAdditionalField

        def test_native_missing(self, module: Any) -> None:
            cfg = OmegaConf.create(module.WithNativeMISSING)
            assert OmegaConf.is_missing(cfg, "num")

        def test_allow_objects(self, module: Any) -> None:
            cfg = OmegaConf.structured(module.Plugin)
            iv = IllegalType()
            with raises(UnsupportedValueType):
                cfg.params = iv
            cfg = OmegaConf.structured(module.Plugin, flags={"allow_objects": True})
            cfg.params = iv
            assert cfg.params == iv

            cfg = OmegaConf.structured(module.Plugin)
            with flag_override(cfg, "allow_objects", True):
                cfg.params = iv
                assert cfg.params == iv

            cfg = OmegaConf.structured({"plugin": module.Plugin})
            pwo = module.Plugin(name="foo", params=iv)
            with raises(UnsupportedValueType):
                cfg.plugin = pwo

            with flag_override(cfg, "allow_objects", True):
                cfg.plugin = pwo
                assert cfg.plugin == pwo



"""
This file compares DictConfig methods with the corresponding
methods of standard python's dict.
The following methods are compared:
    __contains__
    __delitem__
    __eq__
    __getitem__
    __setitem__
    get
    pop
    keys
    values
    items

We have separate test classes for the following cases:
    TestUntypedDictConfig: for DictConfig without a set key_type
    TestPrimitiveTypeDunderMethods: for DictConfig where key_type is primitive
    TestEnumTypeDunderMethods: for DictConfig where key_type is Enum
"""
from copy import deepcopy
from enum import Enum
from typing import Any, Dict, Optional

from pytest import fixture, mark, param, raises

from omegaconf import DictConfig, OmegaConf
from omegaconf.errors import ConfigKeyError, ConfigTypeError, KeyValidationError
from tests import Enum1


@fixture(
    params=[
        "str",
        1,
        3.1415,
        True,
        Enum1.FOO,
    ]
)
def key(request: Any) -> Any:
    """A key to test indexing into DictConfig."""
    return request.param


@fixture
def python_dict(data: Dict[Any, Any]) -> Dict[Any, Any]:
    """Just a standard python dictionary, to be used in comparison with DictConfig."""
    return deepcopy(data)


@fixture(params=[None, False, True])
def struct_mode(request: Any) -> Optional[bool]:
    struct_mode: Optional[bool] = request.param
    return struct_mode


@mark.parametrize(
    "data",
    [
        param({"a": 10}, id="str"),
        param({1: "a"}, id="int"),
        param({123.45: "a"}, id="float"),
        param({True: "a"}, id="bool"),
        param({Enum1.FOO: "foo"}, id="Enum1"),
    ],
)
class TestUntypedDictConfig:
    """Compare DictConfig with python dict in the case where key_type is not set."""

    @fixture
    def cfg(self, python_dict: Any, struct_mode: Optional[bool]) -> DictConfig:
        """Create a DictConfig instance from the given data"""
        cfg: DictConfig = DictConfig(content=python_dict)
        OmegaConf.set_struct(cfg, struct_mode)
        return cfg

    def test__setitem__(
        self, python_dict: Any, cfg: DictConfig, key: Any, struct_mode: Optional[bool]
    ) -> None:
        """Ensure that __setitem__ has same effect on python dict and on DictConfig."""
        if struct_mode and key not in cfg:
            with raises(ConfigKeyError):
                cfg[key] = "sentinel"
        else:
            python_dict[key] = "sentinel"
            cfg[key] = "sentinel"
            assert python_dict == cfg

    def test__getitem__(self, python_dict: Any, cfg: DictConfig, key: Any) -> None:
        """Ensure that __getitem__ has same result with python dict as with DictConfig."""
        try:
            result = python_dict[key]
        except KeyError:
            with raises(ConfigKeyError):
                cfg[key]
        else:
            assert result == cfg[key]

    @mark.parametrize("struct_mode", [False, None])
    def test__delitem__(self, python_dict: Any, cfg: DictConfig, key: Any) -> None:
        """Ensure that __delitem__ has same result with python dict as with DictConfig."""
        try:
            del python_dict[key]
            assert key not in python_dict
        except KeyError:
            with raises(ConfigKeyError):
                del cfg[key]
        else:
            del cfg[key]
            assert key not in cfg

    @mark.parametrize("struct_mode", [True])
    def test__delitem__struct_mode(
        self, python_dict: Any, cfg: DictConfig, key: Any
    ) -> None:
        """Ensure that __delitem__ fails in struct_mode"""
        with raises(ConfigTypeError):
            del cfg[key]

    def test__contains__(self, python_dict: Any, cfg: Any, key: Any) -> None:
        """Ensure that __contains__ has same result with python dict as with DictConfig."""
        assert (key in python_dict) == (key in cfg)

    def test__eq__(self, python_dict: Any, cfg: Any, key: Any) -> None:
        assert python_dict == cfg

    def test_get(self, python_dict: Any, cfg: DictConfig, key: Any) -> None:
        """Ensure that __getitem__ has same result with python dict as with DictConfig."""
        assert python_dict.get(key) == cfg.get(key)

    def test_get_with_default(
        self, python_dict: Any, cfg: DictConfig, key: Any
    ) -> None:
        """Ensure that __getitem__ has same result with python dict as with DictConfig."""
        assert python_dict.get(key, "DEFAULT") == cfg.get(key, "DEFAULT")

    @mark.parametrize("struct_mode", [False, None])
    def test_pop(
        self,
        python_dict: Any,
        cfg: DictConfig,
        key: Any,
    ) -> None:
        """Ensure that pop has same result with python dict as with DictConfig."""
        try:
            result = python_dict.pop(key)
        except KeyError:
            with raises(ConfigKeyError):
                cfg.pop(key)
        else:
            assert result == cfg.pop(key)
        assert python_dict.keys() == cfg.keys()

    @mark.parametrize("struct_mode", [True])
    def test_pop_struct_mode(
        self,
        python_dict: Any,
        cfg: DictConfig,
        key: Any,
    ) -> None:
        """Ensure that pop fails in struct mode."""
        with raises(ConfigTypeError):
            cfg.pop(key)

    @mark.parametrize("struct_mode", [False, None])
    def test_pop_with_default(
        self,
        python_dict: Any,
        cfg: DictConfig,
        key: Any,
    ) -> None:
        """Ensure that pop(..., DEFAULT) has same result with python dict as with DictConfig."""
        assert python_dict.pop(key, "DEFAULT") == cfg.pop(key, "DEFAULT")
        assert python_dict.keys() == cfg.keys()

    @mark.parametrize("struct_mode", [True])
    def test_pop_with_default_struct_mode(
        self,
        python_dict: Any,
        cfg: DictConfig,
        key: Any,
    ) -> None:
        """Ensure that pop(..., DEFAULT) fails in struct mode."""
        with raises(ConfigTypeError):
            cfg.pop(key, "DEFAULT")

    def test_keys(self, python_dict: Any, cfg: Any) -> None:
        assert python_dict.keys() == cfg.keys()

    def test_values(self, python_dict: Any, cfg: Any) -> None:
        assert list(python_dict.values()) == list(cfg.values())

    def test_items(self, python_dict: Any, cfg: Any) -> None:
        assert list(python_dict.items()) == list(cfg.items())


@fixture
def cfg_typed(
    python_dict: Any, cfg_key_type: Any, struct_mode: Optional[bool]
) -> DictConfig:
    """Create a DictConfig instance that has strongly-typed keys"""
    cfg_typed: DictConfig = DictConfig(content=python_dict, key_type=cfg_key_type)
    OmegaConf.set_struct(cfg_typed, struct_mode)
    return cfg_typed


@mark.parametrize(
    "cfg_key_type,data",
    [(str, {"a": 10}), (int, {1: "a"}), (float, {123.45: "a"}), (bool, {True: "a"})],
)
class TestPrimitiveTypeDunderMethods:
    """Compare DictConfig with python dict in the case where key_type is a primitive type."""

    def test__setitem__primitive_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
        struct_mode: Optional[bool],
    ) -> None:
        """When DictConfig keys are strongly typed,
        ensure that __setitem__ has same effect on python dict and on DictConfig."""
        if struct_mode and key not in cfg_typed:
            if isinstance(key, cfg_key_type) or (
                cfg_key_type == bool and key in (0, 1)
            ):
                with raises(ConfigKeyError):
                    cfg_typed[key] = "sentinel"
            else:
                with raises(KeyValidationError):
                    cfg_typed[key] = "sentinel"
        else:
            python_dict[key] = "sentinel"
            if isinstance(key, cfg_key_type) or (
                cfg_key_type == bool and key in (0, 1)
            ):
                cfg_typed[key] = "sentinel"
                assert python_dict == cfg_typed
            else:
                with raises(KeyValidationError):
                    cfg_typed[key] = "sentinel"

    def test__getitem__primitive_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """When Dictconfig keys are strongly typed,
        ensure that __getitem__ has same result with python dict as with DictConfig."""
        try:
            result = python_dict[key]
        except KeyError:
            if isinstance(key, cfg_key_type) or (
                cfg_key_type == bool and key in (0, 1)
            ):
                with raises(ConfigKeyError):
                    cfg_typed[key]
            else:
                with raises(KeyValidationError):
                    cfg_typed[key]
        else:
            assert result == cfg_typed[key]

    @mark.parametrize("struct_mode", [False, None])
    def test__delitem__primitive_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """When Dictconfig keys are strongly typed,
        ensure that __delitem__ has same result with python dict as with DictConfig."""
        try:
            del python_dict[key]
            assert key not in python_dict
        except KeyError:
            if isinstance(key, cfg_key_type) or (
                cfg_key_type == bool and key in (0, 1)
            ):
                with raises(ConfigKeyError):
                    del cfg_typed[key]
            else:
                with raises(KeyValidationError):
                    del cfg_typed[key]
        else:
            del cfg_typed[key]
            assert key not in cfg_typed

    @mark.parametrize("struct_mode", [True])
    def test__delitem__primitive_typed_struct_mode(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure ensure that struct-mode __delitem__ raises ConfigTypeError or KeyValidationError"""
        if isinstance(key, cfg_key_type) or (cfg_key_type == bool and key in (0, 1)):
            with raises(ConfigTypeError):
                del cfg_typed[key]
        else:
            with raises(KeyValidationError):
                del cfg_typed[key]

    def test__contains__primitive_typed(
        self, python_dict: Any, cfg_typed: Any, key: Any
    ) -> None:
        """Ensure that __contains__ has same result with python dict as with DictConfig."""
        assert (key in python_dict) == (key in cfg_typed)

    def test__eq__primitive_typed(
        self, python_dict: Any, cfg_typed: Any, key: Any
    ) -> None:
        assert python_dict == cfg_typed

    def test_get_primitive_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that __getitem__ has same result with python dict as with DictConfig."""
        if isinstance(key, cfg_key_type) or (cfg_key_type == bool and key in (0, 1)):
            assert python_dict.get(key) == cfg_typed.get(key)
        else:
            with raises(KeyValidationError):
                cfg_typed.get(key)

    def test_get_with_default_primitive_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that __getitem__ has same result with python dict as with DictConfig."""
        if isinstance(key, cfg_key_type) or (cfg_key_type == bool and key in (0, 1)):
            assert python_dict.get(key, "DEFAULT") == cfg_typed.get(key, "DEFAULT")
        else:
            with raises(KeyValidationError):
                cfg_typed.get(key, "DEFAULT")

    @mark.parametrize("struct_mode", [False, None])
    def test_pop_primitive_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop has same result with python dict as with DictConfig."""
        if isinstance(key, cfg_key_type) or (cfg_key_type == bool and key in (0, 1)):
            try:
                result = python_dict.pop(key)
            except KeyError:
                with raises(ConfigKeyError):
                    cfg_typed.pop(key)
            else:
                assert result == cfg_typed.pop(key)
            assert python_dict.keys() == cfg_typed.keys()
        else:
            with raises(KeyValidationError):
                cfg_typed.pop(key)

    @mark.parametrize("struct_mode", [True])
    def test_pop_primitive_typed_struct_mode(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop fails in struct mode."""
        with raises(ConfigTypeError):
            cfg_typed.pop(key)

    @mark.parametrize("struct_mode", [False, None])
    def test_pop_with_default_primitive_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop(..., DEFAULT) has same result with python dict as with DictConfig."""
        if isinstance(key, cfg_key_type) or (cfg_key_type == bool and key in (0, 1)):
            assert python_dict.pop(key, "DEFAULT") == cfg_typed.pop(key, "DEFAULT")
            assert python_dict.keys() == cfg_typed.keys()
        else:
            with raises(KeyValidationError):
                cfg_typed.pop(key, "DEFAULT")

    @mark.parametrize("struct_mode", [True])
    def test_pop_with_default_primitive_typed_struct_mode(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop(..., DEFAULT) fails in struct mode"""
        with raises(ConfigTypeError):
            cfg_typed.pop(key)

    def test_keys_primitive_typed(self, python_dict: Any, cfg_typed: Any) -> None:
        assert python_dict.keys() == cfg_typed.keys()

    def test_values_primitive_typed(self, python_dict: Any, cfg_typed: Any) -> None:
        assert list(python_dict.values()) == list(cfg_typed.values())

    def test_items_primitive_typed(self, python_dict: Any, cfg_typed: Any) -> None:
        assert list(python_dict.items()) == list(cfg_typed.items())


@mark.parametrize("cfg_key_type,data", [(Enum1, {Enum1.FOO: "foo"})])
class TestEnumTypeDunderMethods:
    """Compare DictConfig with python dict in the case where key_type is an Enum type."""

    @fixture
    def key_coerced(self, key: Any, cfg_key_type: Any) -> Any:
        """
        This handles key coersion in the special case where DictConfig key_type
        is a subclass of Enum: keys of type `str` or `int` are coerced to `key_type`.
        See https://github.com/omry/omegaconf/pull/484#issuecomment-765772019
        """
        assert issubclass(cfg_key_type, Enum)
        if type(key) == str and key in [e.name for e in cfg_key_type]:
            return cfg_key_type[key]
        elif type(key) == int and key in [e.value for e in cfg_key_type]:
            return cfg_key_type(key)
        else:
            return key

    def test__setitem__enum_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
        struct_mode: Optional[bool],
    ) -> None:
        """When DictConfig keys are strongly typed,
        ensure that __setitem__ has same effect on python dict and on DictConfig."""
        if struct_mode and key_coerced not in cfg_typed:
            if isinstance(key_coerced, cfg_key_type):
                with raises(ConfigKeyError):
                    cfg_typed[key] = "sentinel"
            else:
                with raises(KeyValidationError):
                    cfg_typed[key] = "sentinel"
        else:
            python_dict[key_coerced] = "sentinel"
            if isinstance(key_coerced, cfg_key_type):
                cfg_typed[key] = "sentinel"
                assert python_dict == cfg_typed
            else:
                with raises(KeyValidationError):
                    cfg_typed[key] = "sentinel"

    def test__getitem__enum_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """When Dictconfig keys are strongly typed,
        ensure that __getitem__ has same result with python dict as with DictConfig."""
        try:
            result = python_dict[key_coerced]
        except KeyError:
            if isinstance(key_coerced, cfg_key_type):
                with raises(ConfigKeyError):
                    cfg_typed[key]
            else:
                with raises(KeyValidationError):
                    cfg_typed[key]
        else:
            assert result == cfg_typed[key]

    @mark.parametrize("struct_mode", [False, None])
    def test__delitem__enum_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """When Dictconfig keys are strongly typed,
        ensure that __delitem__ has same result with python dict as with DictConfig."""
        try:
            del python_dict[key_coerced]
            assert key_coerced not in python_dict
        except KeyError:
            if isinstance(key_coerced, cfg_key_type):
                with raises(ConfigKeyError):
                    del cfg_typed[key]
            else:
                with raises(KeyValidationError):
                    del cfg_typed[key]
        else:
            del cfg_typed[key]
            assert key not in cfg_typed

    @mark.parametrize("struct_mode", [True])
    def test__delitem__enum_typed_struct_mode(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that __delitem__ errors in struct mode"""
        if isinstance(key_coerced, cfg_key_type):
            with raises(ConfigTypeError):
                del cfg_typed[key]
        else:
            with raises(KeyValidationError):
                del cfg_typed[key]

    def test__contains__enum_typed(
        self, python_dict: Any, cfg_typed: Any, key: Any, key_coerced: Any
    ) -> None:
        """Ensure that __contains__ has same result with python dict as with DictConfig."""
        assert (key_coerced in python_dict) == (key in cfg_typed)

    def test__eq__enum_typed(self, python_dict: Any, cfg_typed: Any, key: Any) -> None:
        assert python_dict == cfg_typed

    def test_get_enum_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that __getitem__ has same result with python dict as with DictConfig."""
        if isinstance(key_coerced, cfg_key_type):
            assert python_dict.get(key_coerced) == cfg_typed.get(key)
        else:
            with raises(KeyValidationError):
                cfg_typed.get(key)

    def test_get_with_default_enum_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that __getitem__ has same result with python dict as with DictConfig."""
        if isinstance(key_coerced, cfg_key_type):
            assert python_dict.get(key_coerced, "DEFAULT") == cfg_typed.get(
                key, "DEFAULT"
            )
        else:
            with raises(KeyValidationError):
                cfg_typed.get(key, "DEFAULT")

    @mark.parametrize("struct_mode", [False, None])
    def test_pop_enum_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop has same result with python dict as with DictConfig."""
        if isinstance(key_coerced, cfg_key_type):
            try:
                result = python_dict.pop(key_coerced)
            except KeyError:
                with raises(ConfigKeyError):
                    cfg_typed.pop(key)
            else:
                assert result == cfg_typed.pop(key)
            assert python_dict.keys() == cfg_typed.keys()
        else:
            with raises(KeyValidationError):
                cfg_typed.pop(key)

    @mark.parametrize("struct_mode", [True])
    def test_pop_enum_typed_struct_mode(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop fails in struct mode"""
        with raises(ConfigTypeError):
            cfg_typed.pop(key)

    @mark.parametrize("struct_mode", [False, None])
    def test_pop_with_default_enum_typed(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop(..., DEFAULT) has same result with python dict as with DictConfig."""
        if isinstance(key_coerced, cfg_key_type):
            assert python_dict.pop(key_coerced, "DEFAULT") == cfg_typed.pop(
                key, "DEFAULT"
            )
            assert python_dict.keys() == cfg_typed.keys()
        else:
            with raises(KeyValidationError):
                cfg_typed.pop(key, "DEFAULT")

    @mark.parametrize("struct_mode", [True])
    def test_pop_with_default_enum_typed_struct_mode(
        self,
        python_dict: Any,
        cfg_typed: DictConfig,
        key: Any,
        key_coerced: Any,
        cfg_key_type: Any,
    ) -> None:
        """Ensure that pop(..., DEFAULT) errors in struct mode"""
        with raises(ConfigTypeError):
            cfg_typed.pop(key)

    def test_keys_enum_typed(self, python_dict: Any, cfg_typed: Any) -> None:
        assert python_dict.keys() == cfg_typed.keys()

    def test_values_enum_typed(self, python_dict: Any, cfg_typed: Any) -> None:
        assert list(python_dict.values()) == list(cfg_typed.values())

    def test_items_enum_typed(self, python_dict: Any, cfg_typed: Any) -> None:
        assert list(python_dict.items()) == list(cfg_typed.items())



"""Testing for OmegaConf"""
import re
import sys
from textwrap import dedent
from typing import Any, Dict, List, Optional

import yaml
from pytest import mark, param, raises

from omegaconf import DictConfig, ListConfig, OmegaConf
from omegaconf.errors import UnsupportedValueType
from tests import ConcretePlugin, IllegalType, NonCopyableIllegalType, Plugin


@mark.parametrize(
    "input_,expected",
    [
        # No content
        (None, None),
        # empty
        ({}, {}),
        # simple value
        ("hello", {"hello": None}),
        # simple key:value"
        ("hello: world", {"hello": "world"}),
        ({"hello": {"a": 2}}, {"hello": {"a": 2}}),
        # empty input
        ("", {}),
        # list value
        ([1, 2], [1, 2]),
        # For simplicity, tuples are converted to lists.
        ((1, 2), [1, 2]),
        # dict 1
        ({"a": 2, "b": 10}, {"a": 2, "b": 10}),
        # dict 2
        (dict(a=2, b=10), dict(a=2, b=10)),
        # nested dict
        (
            {"a": 2, "b": {"c": {"f": 1}, "d": {}}},
            {"a": 2, "b": {"c": {"f": 1}, "d": {}}},
        ),
        ({"a": None}, {"a": None}),
        ({"foo": "${missing}"}, {"foo": "${missing}"}),
        (OmegaConf.create({"foo": "${missing}"}), {"foo": "${missing}"}),
        (OmegaConf.create(), {}),
        (OmegaConf.create({}), {}),
        (OmegaConf.create([]), []),
        (OmegaConf.create({"foo": OmegaConf.create([])}), {"foo": []}),
        (OmegaConf.create([OmegaConf.create({})]), [{}]),
    ],
)
def test_create_value(input_: Any, expected: Any) -> None:
    assert OmegaConf.create(input_) == expected


@mark.parametrize(
    "input_",
    [
        # top level dict
        {"x": IllegalType()},
        {"x": {"y": IllegalType()}},
        {"x": [IllegalType()]},
        # top level list
        [IllegalType()],
        [[IllegalType()]],
        [{"x": IllegalType()}],
        [{"x": [IllegalType()]}],
    ],
)
def test_create_allow_objects(input_: Any) -> None:
    # test creating from a primitive container
    cfg = OmegaConf.create(input_, flags={"allow_objects": True})
    assert cfg == input_

    # test creating from an OmegaConf object, inheriting the allow_objects flag
    cfg = OmegaConf.create(cfg)
    assert cfg == input_

    # test creating from an OmegaConf object
    cfg = OmegaConf.create(cfg, flags={"allow_objects": True})
    assert cfg == input_


@mark.parametrize(
    "input_",
    [
        # top level dict
        {"x": NonCopyableIllegalType()},
        {"x": {"y": NonCopyableIllegalType()}},
        {"x": [NonCopyableIllegalType()]},
        # top level list
        [NonCopyableIllegalType()],
        [[NonCopyableIllegalType()]],
        [{"x": NonCopyableIllegalType()}],
        [{"x": [NonCopyableIllegalType()]}],
    ],
)
def test_create_allow_objects_non_copyable(input_: Any) -> None:
    # test creating from a primitive container
    cfg = OmegaConf.create(input_, flags={"allow_objects": True})
    assert cfg == input_

    # test creating from an OmegaConf object, inheriting the allow_objects flag
    cfg = OmegaConf.create(cfg)
    assert cfg == input_

    # test creating from an OmegaConf object
    cfg = OmegaConf.create(cfg, flags={"allow_objects": True})
    assert cfg == input_


@mark.parametrize(
    "input_",
    [
        param({"foo": "bar"}, id="dict"),
        param([1, 2, 3], id="list"),
    ],
)
def test_create_flags_overriding(input_: Any) -> Any:
    cfg = OmegaConf.create(input_)
    OmegaConf.set_struct(cfg, True)

    # by default flags are inherited
    cfg2 = OmegaConf.create(cfg)
    assert OmegaConf.is_struct(cfg2)
    assert not OmegaConf.is_readonly(cfg2)

    # but specified flags are replacing all of the flags (even those that are not specified)
    cfg2 = OmegaConf.create(cfg, flags={"readonly": True})
    assert not OmegaConf.is_struct(cfg2)
    assert OmegaConf.is_readonly(cfg2)


def test_create_from_cli() -> None:
    sys.argv = ["program.py", "a=1", "b.c=2"]
    c = OmegaConf.from_cli()
    assert {"a": 1, "b": {"c": 2}} == c


def test_cli_passing() -> None:
    args_list = ["a=1", "b.c=2"]
    c = OmegaConf.from_cli(args_list)
    assert {"a": 1, "b": {"c": 2}} == c


@mark.parametrize(
    "input_,expected",
    [
        # simple
        (["a=1", "b.c=2"], dict(a=1, b=dict(c=2))),
        # string
        (["a=hello", "b=world"], dict(a="hello", b="world")),
        # date-formatted string
        (["my_date=2019-12-11"], dict(my_date="2019-12-11")),
    ],
)
def test_dotlist(input_: List[str], expected: Dict[str, Any]) -> None:
    c = OmegaConf.from_dotlist(input_)
    assert c == expected


def test_create_list_with_illegal_value_idx0() -> None:
    with raises(UnsupportedValueType, match=re.escape("key: [0]")):
        OmegaConf.create([IllegalType()])


def test_create_list_with_illegal_value_idx1() -> None:
    lst = [1, IllegalType(), 3]
    with raises(UnsupportedValueType, match=re.escape("key: [1]")):
        OmegaConf.create(lst)


def test_create_dict_with_illegal_value() -> None:
    with raises(UnsupportedValueType, match=re.escape("key: a")):
        OmegaConf.create({"a": IllegalType()})


def test_create_nested_dict_with_illegal_value() -> None:
    with raises(ValueError, match=re.escape("key: a.b")):
        OmegaConf.create({"a": {"b": IllegalType()}})


def test_create_from_oc() -> None:
    c = OmegaConf.create(
        {"a": OmegaConf.create([1, 2, 3]), "b": OmegaConf.create({"c": 10})}
    )
    assert c == {"a": [1, 2, 3], "b": {"c": 10}}


def test_create_from_oc_with_flags() -> None:
    c1 = OmegaConf.create({"foo": "bar"})
    OmegaConf.set_struct(c1, True)
    c2 = OmegaConf.create(c1)
    assert c1 == c2
    assert c1._metadata.flags == c2._metadata.flags


def test_create_from_dictconfig_preserves_metadata() -> None:
    cfg1 = DictConfig(ref_type=Plugin, is_optional=False, content=ConcretePlugin)
    OmegaConf.set_struct(cfg1, True)
    OmegaConf.set_readonly(cfg1, True)
    cfg2 = OmegaConf.create(cfg1)
    assert cfg1 == cfg2
    assert cfg1._metadata == cfg2._metadata


def test_create_from_listconfig_preserves_metadata() -> None:
    cfg1 = ListConfig(element_type=int, is_optional=False, content=[1, 2, 3])
    OmegaConf.set_struct(cfg1, True)
    OmegaConf.set_readonly(cfg1, True)
    cfg2 = OmegaConf.create(cfg1)
    assert cfg1 == cfg2
    assert cfg1._metadata == cfg2._metadata


@mark.parametrize("node", [({"bar": 10}), ([1, 2, 3])])
def test_create_node_parent_retained_on_create(node: Any) -> None:
    cfg1 = OmegaConf.create({"foo": node})
    cfg2 = OmegaConf.create({"zonk": cfg1.foo})
    assert cfg2 == {"zonk": node}
    assert cfg1.foo._get_parent() == cfg1
    assert cfg1.foo._get_parent() is cfg1


@mark.parametrize("node", [({"bar": 10}), ([1, 2, 3])])
def test_create_node_parent_retained_on_assign(node: Any) -> None:
    cfg1 = OmegaConf.create({"foo": node})
    cfg2 = OmegaConf.create()
    cfg2.zonk = cfg1.foo
    assert cfg1.foo._get_parent() is cfg1
    assert cfg2.zonk._get_parent() is cfg2


@mark.parametrize(
    "node",
    [
        {"a": 0},
        DictConfig({"a": 0}),
    ],
)
def test_dict_assignment_deepcopy_semantics(node: Any) -> None:
    cfg = OmegaConf.create()
    cfg.foo = node
    node["a"] = 1
    assert cfg.foo.a == 0


@mark.parametrize(
    "node",
    [
        [1, 2],
        ListConfig([1, 2]),
    ],
)
def test_list_assignment_deepcopy_semantics(node: Any) -> None:
    cfg = OmegaConf.create()
    cfg.foo = node
    node[1] = 10
    assert cfg.foo[1] == 2


@mark.parametrize("d", [{"a": {"b": 10}}, {"a": {"b": {"c": 10}}}])
def test_assign_does_not_modify_src_config(d: Any) -> None:
    cfg1 = OmegaConf.create(d)
    cfg2 = OmegaConf.create({})
    cfg2.a = cfg1.a
    assert cfg1 == d
    assert cfg2 == d

    assert cfg1.a._get_parent() is cfg1
    assert cfg1.a._get_node("b")._get_parent() is cfg1.a

    assert cfg2.a._get_parent() is cfg2
    assert cfg2.a._get_node("b")._get_parent() is cfg2.a


def test_create_unmodified_loader() -> None:
    cfg = OmegaConf.create("gitrev: 100e100")
    yaml_cfg = yaml.load("gitrev: 100e100", Loader=yaml.loader.SafeLoader)
    assert cfg.gitrev == 1e102
    assert yaml_cfg["gitrev"] == "100e100"


def test_create_untyped_list() -> None:
    from omegaconf._utils import get_ref_type

    cfg = ListConfig(ref_type=List, content=[])
    assert get_ref_type(cfg) == Optional[List]


def test_create_untyped_dict() -> None:
    from omegaconf._utils import get_ref_type

    cfg = DictConfig(ref_type=Dict, content={})
    assert get_ref_type(cfg) == Optional[Dict]


@mark.parametrize(
    "input_",
    [
        dedent(
            """\
                a:
                  b: 1
                  c: 2
                  b: 3
                """
        ),
        dedent(
            """\
                a:
                  b: 1
                a:
                  b: 2
                """
        ),
    ],
)
def test_yaml_duplicate_keys(input_: str) -> None:
    with raises(yaml.constructor.ConstructorError):
        OmegaConf.create(input_)


def test_yaml_merge() -> None:
    cfg = OmegaConf.create(
        dedent(
            """\
                a: &A
                    x: 1
                b: &B
                    y: 2
                c:
                    <<: *A
                    <<: *B
                    x: 3
                    z: 1
                """
        )
    )
    assert cfg == {"a": {"x": 1}, "b": {"y": 2}, "c": {"x": 3, "y": 2, "z": 1}}


@mark.parametrize(
    "data",
    [
        param("", id="empty"),
        param("hello", id="name_only"),
        param("a: b", id="dictconfig"),
        param("- a", id="listconfig"),
    ],
)
def test_create_from_str_check_parent(data: str) -> None:
    parent = OmegaConf.create({})
    cfg = OmegaConf.create(data, parent=parent)
    assert cfg._get_parent() is parent



import re
from typing import Any

from pytest import mark, param, raises, warns

from omegaconf import (
    MISSING,
    BooleanNode,
    DictConfig,
    EnumNode,
    FloatNode,
    IntegerNode,
    ListConfig,
    MissingMandatoryValue,
    OmegaConf,
    StringNode,
)
from omegaconf._utils import _is_none, nullcontext
from omegaconf.errors import (
    ConfigKeyError,
    InterpolationKeyError,
    InterpolationToMissingValueError,
    UnsupportedInterpolationType,
)
from tests import Color, ConcretePlugin, IllegalType, StructuredWithMissing


@mark.parametrize(
    "cfg, key, expected_is_missing, expectation",
    [
        ({}, "foo", False, raises(ConfigKeyError)),
        ({"foo": True}, "foo", False, nullcontext()),
        ({"foo": "${no_such_key}"}, "foo", False, raises(InterpolationKeyError)),
        ({"foo": MISSING}, "foo", True, raises(MissingMandatoryValue)),
        param(
            {"foo": "${bar}", "bar": DictConfig(content=MISSING)},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
            id="missing_interpolated_dict",
        ),
        param(
            {"foo": ListConfig(content="???")},
            "foo",
            True,
            raises(MissingMandatoryValue),
            id="missing_list",
        ),
        param(
            {"foo": DictConfig(content="???")},
            "foo",
            True,
            raises(MissingMandatoryValue),
            id="missing_dict",
        ),
        param(
            {"foo": ListConfig(content="${missing}"), "missing": "???"},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
            id="missing_list_interpolation",
        ),
        param(
            {"foo": DictConfig(content="${missing}"), "missing": "???"},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
            id="missing_dict_interpolation",
        ),
        (
            {"foo": "${bar}", "bar": MISSING},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            {"foo": "foo_${bar}", "bar": MISSING},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            {"foo": "${unknown_resolver:foo}"},
            "foo",
            False,
            raises(UnsupportedInterpolationType),
        ),
        ({"foo": StringNode(value="???")}, "foo", True, raises(MissingMandatoryValue)),
        (
            {"foo": StringNode(value="???"), "inter": "${foo}"},
            "inter",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (StructuredWithMissing, "num", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_num", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "dict", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_dict", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "list", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_list", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "user", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_user", True, raises(MissingMandatoryValue)),
        (
            StructuredWithMissing,
            "inter_user",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            StructuredWithMissing,
            "inter_opt_user",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            StructuredWithMissing,
            "inter_num",
            False,
            raises(InterpolationToMissingValueError),
        ),
    ],
)
def test_is_missing(
    cfg: Any, key: str, expected_is_missing: bool, expectation: Any
) -> None:
    cfg = OmegaConf.create(cfg)
    with expectation:
        cfg[key]

    assert OmegaConf.is_missing(cfg, key) == expected_is_missing
    OmegaConf.set_struct(cfg, True)
    assert OmegaConf.is_missing(cfg, key) == expected_is_missing
    OmegaConf.set_readonly(cfg, True)
    assert OmegaConf.is_missing(cfg, key) == expected_is_missing


def test_is_missing_resets() -> None:
    cfg = OmegaConf.structured(StructuredWithMissing)
    assert OmegaConf.is_missing(cfg, "dict")
    cfg.dict = {}
    assert not OmegaConf.is_missing(cfg, "dict")

    assert OmegaConf.is_missing(cfg, "list")
    cfg.list = [1, 2, 3]
    assert not OmegaConf.is_missing(cfg, "list")
    cfg.list = "???"
    assert OmegaConf.is_missing(cfg, "list")


@mark.parametrize(
    "cfg, expected",
    [
        (None, False),
        ({}, False),
        ([], False),
        ("aa", False),
        (10, False),
        (True, False),
        (bool, False),
        (StringNode("foo"), False),
        (ConcretePlugin, False),
        (ConcretePlugin(), False),
        (OmegaConf.create({}), True),
        (OmegaConf.create([]), True),
        (OmegaConf.structured(ConcretePlugin), True),
        (OmegaConf.structured(ConcretePlugin()), True),
    ],
)
def test_is_config(cfg: Any, expected: bool) -> None:
    assert OmegaConf.is_config(cfg) == expected


@mark.parametrize(
    "cfg, expected",
    [
        (None, False),
        ({}, False),
        ([], False),
        ("aa", False),
        (10, False),
        (True, False),
        (bool, False),
        (StringNode("foo"), False),
        (ConcretePlugin, False),
        (ConcretePlugin(), False),
        (OmegaConf.create({}), False),
        (OmegaConf.create([]), True),
        (OmegaConf.structured(ConcretePlugin), False),
        (OmegaConf.structured(ConcretePlugin()), False),
    ],
)
def test_is_list(cfg: Any, expected: bool) -> None:
    assert OmegaConf.is_list(cfg) == expected


@mark.parametrize(
    "cfg, expected",
    [
        (None, False),
        ({}, False),
        ([], False),
        ("aa", False),
        (10, False),
        (True, False),
        (bool, False),
        (StringNode("foo"), False),
        (ConcretePlugin, False),
        (ConcretePlugin(), False),
        (OmegaConf.create({}), True),
        (OmegaConf.create([]), False),
        (OmegaConf.structured(ConcretePlugin), True),
        (OmegaConf.structured(ConcretePlugin()), True),
    ],
)
def test_is_dict(cfg: Any, expected: bool) -> None:
    assert OmegaConf.is_dict(cfg) == expected


def test_is_optional_deprecation() -> None:
    with warns(UserWarning, match=re.escape("`OmegaConf.is_optional()` is deprecated")):
        OmegaConf.is_optional("xxx")


def test_coverage_for_deprecated_OmegaConf_is_optional() -> None:
    cfg = OmegaConf.create({"node": 123})
    with warns(UserWarning):
        assert OmegaConf.is_optional(cfg)
    with warns(UserWarning):
        assert OmegaConf.is_optional(cfg, "node")
    with warns(UserWarning):
        assert OmegaConf.is_optional("not_a_node")


@mark.parametrize("is_none", [True, False])
@mark.parametrize(
    "fac",
    [
        (lambda none: StringNode(value="foo" if not none else None, is_optional=True)),
        (lambda none: IntegerNode(value=10 if not none else None, is_optional=True)),
        (lambda none: FloatNode(value=10 if not none else None, is_optional=True)),
        (lambda none: BooleanNode(value=True if not none else None, is_optional=True)),
        (
            lambda none: EnumNode(
                enum_type=Color,
                value=Color.RED if not none else None,
                is_optional=True,
            )
        ),
        (
            lambda none: ListConfig(
                content=[1, 2, 3] if not none else None, is_optional=True
            )
        ),
        (
            lambda none: DictConfig(
                content={"foo": "bar"} if not none else None, is_optional=True
            )
        ),
        (
            lambda none: DictConfig(
                ref_type=ConcretePlugin,
                content=ConcretePlugin() if not none else None,
                is_optional=True,
            )
        ),
    ],
)
def test_is_none(fac: Any, is_none: bool) -> None:
    obj = fac(is_none)
    assert obj._is_none() == is_none

    cfg = OmegaConf.create({"node": obj})
    assert (cfg.node is None) == is_none


@mark.parametrize(
    ("cfg", "key", "is_none"),
    [
        ({"foo": "${none}", "none": None}, "foo", True),
        ({"foo": "${missing_node}"}, "foo", False),
        ({"foo": "${missing_resolver:}"}, "foo", False),
        ({"foo": "${oc.decode:'[1, '}"}, "foo", False),
    ],
)
def test_is_none_interpolation(cfg: Any, key: str, is_none: bool) -> None:
    cfg = OmegaConf.create(cfg)
    with warns(UserWarning):
        assert OmegaConf.is_none(cfg, key) == is_none
    check = _is_none(
        cfg._get_node(key), resolve=True, throw_on_resolution_failure=False
    )
    assert check == is_none


def test_is_none_invalid_node() -> None:
    cfg = OmegaConf.create({})
    with warns(UserWarning):
        assert OmegaConf.is_none(cfg, "invalid")


@mark.parametrize(
    "fac",
    [
        (
            lambda inter: StringNode(
                value="foo" if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: IntegerNode(
                value=10 if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: FloatNode(
                value=10 if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: BooleanNode(
                value=True if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: EnumNode(
                enum_type=Color,
                value=Color.RED if inter is None else inter,
                is_optional=True,
            )
        ),
        (
            lambda inter: ListConfig(
                content=[1, 2, 3] if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: DictConfig(
                content={"foo": "bar"} if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: DictConfig(
                ref_type=ConcretePlugin,
                content=ConcretePlugin() if inter is None else inter,
                is_optional=True,
            )
        ),
    ],
    ids=[
        "StringNode",
        "IntegerNode",
        "FloatNode",
        "BooleanNode",
        "EnumNode",
        "ListConfig",
        "DictConfig",
        "ConcretePlugin",
    ],
)
def test_is_interpolation(fac: Any) -> Any:
    obj = fac(inter=None)
    assert not OmegaConf.is_interpolation(obj)
    cfg = OmegaConf.create({"node": obj})
    assert not OmegaConf.is_interpolation(cfg, "node")

    assert not OmegaConf.is_interpolation(cfg, "missing")

    for inter in ["${foo}", "http://${url}", "${resolver:value}"]:
        obj = fac(inter=inter)
        assert OmegaConf.is_interpolation(obj)
        cfg = OmegaConf.create({"node": obj})
        assert OmegaConf.is_interpolation(cfg, "node")


@mark.parametrize(
    "cfg, type_",
    [
        ({"foo": 10}, int),
        ({"foo": 10.0}, float),
        ({"foo": True}, bool),
        ({"foo": "bar"}, str),
        ({"foo": None}, type(None)),
        ({"foo": ConcretePlugin()}, ConcretePlugin),
        ({"foo": ConcretePlugin}, ConcretePlugin),
        ({"foo": {}}, dict),
        ({"foo": OmegaConf.create()}, dict),
        ({"foo": []}, list),
        ({"foo": OmegaConf.create([])}, list),
    ],
)
def test_get_type(cfg: Any, type_: Any) -> None:
    cfg = OmegaConf.create(cfg)
    assert OmegaConf.get_type(cfg, "foo") == type_


@mark.parametrize(
    "obj, type_",
    [
        (10, int),
        (10.0, float),
        (True, bool),
        ("foo", str),
        (DictConfig(content={}), dict),
        (ListConfig(content=[]), list),
        (IllegalType, IllegalType),
        (IllegalType(), IllegalType),
    ],
)
def test_get_type_on_raw(obj: Any, type_: Any) -> None:
    assert OmegaConf.get_type(obj) == type_


def test_is_issubclass() -> None:
    cfg = OmegaConf.structured(ConcretePlugin)
    t = OmegaConf.get_type(cfg)
    assert t is not None and issubclass(t, ConcretePlugin)


@mark.parametrize(
    ("cfg", "expected"),
    [
        # dict
        param(OmegaConf.create(), {}, id="dict"),
        param(OmegaConf.create({"a": 10, "b": "${a}"}), {"a": 10, "b": 10}, id="dict"),
        param(
            OmegaConf.create({"a": 10, "b": {"a": "${a}"}}),
            {"a": 10, "b": {"a": 10}},
            id="dict",
        ),
        param(
            OmegaConf.create({"a": "${b.a}", "b": {"a": 10}}),
            {"a": 10, "b": {"a": 10}},
            id="dict",
        ),
        param(OmegaConf.create({"a": "???"}), {"a": "???"}, id="dict:missing"),
        param(
            OmegaConf.create({"a": "???", "b": "${a}"}),
            {"a": "???", "b": "???"},
            id="dict:missing",
        ),
        param(
            OmegaConf.create({"a": 10, "b": "a_${a}"}),
            {"a": 10, "b": "a_10"},
            id="dict:str_inter",
        ),
        # This seems like a reasonable resolution for a string interpolation pointing to a missing node:
        param(
            OmegaConf.create({"a": "???", "b": "a_${a}"}),
            {"a": "???", "b": "???"},
            id="dict:str_inter_missing",
        ),
        param(
            DictConfig("${a}", parent=OmegaConf.create({"a": {"b": 10}})),
            {"b": 10},
            id="inter_dict",
        ),
        param(
            DictConfig("${a}", parent=OmegaConf.create({"a": "???"})),
            "???",
            id="inter_dict_to_missing",
        ),
        param(
            OmegaConf.create({"x": "${y}", "y": {"z": "${foo}"}, "foo": 0}),
            {"x": {"z": 0}, "y": {"z": 0}, "foo": 0},
            id="dict_nested_interpolation",
        ),
        param(DictConfig(None), None, id="none_dict"),
        # comparing to ??? because to_container returns it.
        param(DictConfig("???"), "???", id="missing_dict"),
        # lists
        param(OmegaConf.create([]), [], id="list"),
        param(OmegaConf.create([10, "${0}"]), [10, 10], id="list"),
        param(OmegaConf.create(["???"]), ["???"], id="list:missing"),
        param(OmegaConf.create(["${1}", "???"]), ["???", "???"], id="list:missing"),
        param(
            ListConfig("${a}", parent=OmegaConf.create({"a": [1, 2]})),
            [1, 2],
            id="inter_list",
        ),
        param(
            ListConfig("${a}", parent=OmegaConf.create({"a": "???"})),
            "???",
            id="inter_list_to_missing",
        ),
        param(ListConfig(None), None, id="none_list"),
        # comparing to ??? because to_container returns it.
        param(ListConfig("???"), "???", id="missing_list"),
        # Tests for cases where an AnyNode with interpolation is promoted to a DictConfig/ListConfig node
        param(
            OmegaConf.create({"a": "${z}", "z": {"y": 1}}),
            {"a": {"y": 1}, "z": {"y": 1}},
            id="any_in_dict_to_dict",
        ),
        param(
            OmegaConf.create({"a": "${z}", "z": [1, 2]}),
            {"a": [1, 2], "z": [1, 2]},
            id="any_in_dict_to_list",
        ),
        param(
            OmegaConf.create(["${1}", {"z": {"y": 1}}]),
            [{"z": {"y": 1}}, {"z": {"y": 1}}],
            id="any_in_list_to_dict",
        ),
        param(
            OmegaConf.create(["${1}", [1, 2]]),
            [[1, 2], [1, 2]],
            id="any_in_list_to_list",
        ),
    ],
)
def test_resolve(cfg: Any, expected: Any) -> None:
    OmegaConf.resolve(cfg)
    # convert output to plain dict to avoid smart OmegaConf eq logic
    resolved_plain = OmegaConf.to_container(cfg, resolve=False)
    assert resolved_plain == expected


def test_resolve_invalid_input() -> None:
    with raises(ValueError):
        OmegaConf.resolve("aaa")  # type: ignore



# based on https://github.com/fabioz/PyDev.Debugger/tree/main/pydevd_plugins/extensions
import os
import sys
from typing import Any, Dict

from _pydevd_bundle.pydevd_extension_api import (  # type: ignore
    StrPresentationProvider,
    TypeResolveProvider,
)

DEBUG = False


def print_debug(msg: str) -> None:  # pragma: no cover
    if DEBUG:
        print(msg)


def find_mod_attr(mod_name: str, attr: str) -> Any:
    mod = sys.modules.get(mod_name)
    return getattr(mod, attr, None)


class OmegaConfDeveloperResolver(object):
    def can_provide(self, type_object: Any, type_name: str) -> bool:
        Node = find_mod_attr("omegaconf", "Node")
        return Node is not None and issubclass(type_object, Node)

    def resolve(self, obj: Any, attribute: str) -> Any:
        return getattr(obj, attribute)

    def get_dictionary(self, obj: Any) -> Any:
        return obj.__dict__


class OmegaConfUserResolver(StrPresentationProvider):  # type: ignore
    def __init__(self) -> None:
        self.Node = find_mod_attr("omegaconf", "Node")
        self.ValueNode = find_mod_attr("omegaconf", "ValueNode")
        self.ListConfig = find_mod_attr("omegaconf", "ListConfig")
        self.DictConfig = find_mod_attr("omegaconf", "DictConfig")
        self.InterpolationResolutionError = find_mod_attr(
            "omegaconf.errors", "InterpolationResolutionError"
        )

    def can_provide(self, type_object: Any, type_name: str) -> bool:
        return self.Node is not None and issubclass(type_object, self.Node)

    def resolve(self, obj: Any, attribute: Any) -> Any:
        if isinstance(obj, self.ListConfig) and isinstance(attribute, str):
            attribute = int(attribute)

        if isinstance(obj, self.Node):
            obj = obj._dereference_node()

        val = obj.__dict__["_content"][attribute]

        print_debug(
            f"resolving {obj} ({type(obj).__name__}), {attribute} -> {val} ({type(val).__name__})"
        )

        return val

    def _is_simple_value(self, val: Any) -> bool:
        return (
            isinstance(val, self.ValueNode)
            and not val._is_none()
            and not val._is_missing()
            and not val._is_interpolation()
        )

    def get_dictionary(self, obj: Any) -> Dict[str, Any]:
        d = self._get_dictionary(obj)
        print_debug(f"get_dictionary {obj}, ({type(obj).__name__}) -> {d}")
        return d

    def _get_dictionary(self, obj: Any) -> Dict[str, Any]:
        if isinstance(obj, self.Node):
            obj = obj._maybe_dereference_node()
            if obj is None or obj._is_none() or obj._is_missing():
                return {}

        if isinstance(obj, self.DictConfig):
            d = {}
            for k, v in obj.__dict__["_content"].items():
                if self._is_simple_value(v):
                    v = v._value()
                d[k] = v
        elif isinstance(obj, self.ListConfig):
            d = {}
            for idx, v in enumerate(obj.__dict__["_content"]):
                if self._is_simple_value(v):
                    v = v._value()
                d[str(idx)] = v
        else:
            d = {}

        return d

    def get_str(self, val: Any) -> str:
        if val._is_missing():
            return "??? <MISSING>"
        if val._is_interpolation():
            try:
                dr = val._dereference_node()
            except self.InterpolationResolutionError as e:
                dr = f"ERR: {e}"
            return f"{val._value()} -> {dr}"
        else:
            return f"{val}"


# OC_PYDEVD_RESOLVER env can take:
#  DISABLE: Do not install a pydevd resolver
#  USER: Install a resolver for OmegaConf users (default)
#  DEV: Install a resolver for OmegaConf developers. Shows underlying data-model in the debugger.
resolver = os.environ.get("OC_PYDEVD_RESOLVER", "USER").upper()
if resolver != "DISABLE":  # pragma: no cover
    if resolver == "USER":
        TypeResolveProvider.register(OmegaConfUserResolver)
    elif resolver == "DEV":
        TypeResolveProvider.register(OmegaConfDeveloperResolver)
    else:
        sys.stderr.write(
            f"OmegaConf pydev plugin: Not installing. Unknown mode {resolver}. Supported one of [USER, DEV, DISABLE]\n"
        )



import re
from typing import Any

from pytest import mark, param, raises, warns

from omegaconf import OmegaConf
from omegaconf.errors import InterpolationResolutionError


@mark.parametrize(
    ("cfg", "key", "expected_value", "expected_warning"),
    [
        param(
            {"a": 10, "b": "${oc.deprecated: a}"},
            "b",
            10,
            "'b' is deprecated. Change your code and config to use 'a'",
            id="value",
        ),
        param(
            {"a": 10, "b": "${oc.deprecated: a, '$OLD_KEY is deprecated'}"},
            "b",
            10,
            "b is deprecated",
            id="value-custom-message",
        ),
        param(
            {
                "a": 10,
                "b": "${oc.deprecated: a, ${warning}}",
                "warning": "$OLD_KEY is bad, $NEW_KEY is good",
            },
            "b",
            10,
            "b is bad, a is good",
            id="value-custom-message-config-variable",
        ),
        param(
            {"a": {"b": 10}, "b": "${oc.deprecated: a}"},
            "b",
            OmegaConf.create({"b": 10}),
            "'b' is deprecated. Change your code and config to use 'a'",
            id="dict",
        ),
        param(
            {"a": {"b": 10}, "b": "${oc.deprecated: a}"},
            "b.b",
            10,
            "'b' is deprecated. Change your code and config to use 'a'",
            id="dict_value",
        ),
        param(
            {"a": [0, 1], "b": "${oc.deprecated: a}"},
            "b",
            OmegaConf.create([0, 1]),
            "'b' is deprecated. Change your code and config to use 'a'",
            id="list",
        ),
        param(
            {"a": [0, 1], "b": "${oc.deprecated: a}"},
            "b[1]",
            1,
            "'b' is deprecated. Change your code and config to use 'a'",
            id="list_value",
        ),
    ],
)
def test_deprecated(
    cfg: Any, key: str, expected_value: Any, expected_warning: str
) -> None:
    cfg = OmegaConf.create(cfg)
    with warns(UserWarning, match=re.escape(expected_warning)):
        value = OmegaConf.select(cfg, key)
    assert value == expected_value
    assert type(value) == type(expected_value)


@mark.parametrize(
    ("cfg", "error"),
    [
        param(
            {"a": "${oc.deprecated: z}"},
            "ConfigKeyError raised while resolving interpolation:"
            " In oc.deprecated resolver at 'a': Key not found: 'z'",
            id="target_not_found",
        ),
        param(
            {"a": "${oc.deprecated: 111111}"},
            "TypeError raised while resolving interpolation: oc.deprecated:"
            " interpolation key type is not a string (int)",
            id="invalid_key_type",
        ),
        param(
            {"a": "${oc.deprecated: b, 1000}", "b": 10},
            "TypeError raised while resolving interpolation: oc.deprecated:"
            " interpolation message type is not a string (int)",
            id="invalid_message_type",
        ),
    ],
)
def test_deprecated_target_not_found(cfg: Any, error: str) -> None:
    cfg = OmegaConf.create(cfg)
    with raises(
        InterpolationResolutionError,
        match=re.escape(error),
    ):
        cfg.a



# `from __future__` has to be the very first thing in a module
# otherwise a syntax error is raised
from __future__ import annotations  # type: ignore # noqa  # Python 3.6 linters complain

from dataclasses import dataclass, fields
from enum import Enum

from pytest import raises

from omegaconf import OmegaConf, ValidationError


class Height(Enum):
    SHORT = 0
    TALL = 1


@dataclass
class SimpleTypes:
    num: int = 10
    pi: float = 3.1415
    is_awesome: bool = True
    height: "Height" = Height.SHORT  # test forward ref
    description: str = "text"


def simple_types_class() -> None:
    # confirm that the type annotations are in fact stored as strings
    # i.e., that the `from future` import worked
    num_field = fields(SimpleTypes)[0]
    assert isinstance(num_field.type, str)
    assert num_field.type == "int"

    conf = OmegaConf.structured(SimpleTypes)
    assert conf.num == 10
    assert conf.pi == 3.1415
    assert conf.is_awesome is True
    assert conf.height == Height.SHORT
    assert conf.description == "text"


def conversions() -> None:
    conf: SimpleTypes = OmegaConf.structured(SimpleTypes)
    conf.num = 20

    conf.num = "20"  # type: ignore
    assert conf.num == 20

    with raises(ValidationError):
        # ValidationError: "one" cannot be converted to an integer
        conf.num = "one"  # type: ignore



import copy
from typing import Any

from pytest import fixture

from omegaconf import OmegaConf
from omegaconf.basecontainer import BaseContainer


@fixture(scope="function")
def restore_resolvers() -> Any:
    """
    A fixture to restore singletons state after this the function.
    This is useful for functions that are making a one-off change to singlestons that should not effect
    other tests
    """
    state = copy.deepcopy(BaseContainer._resolvers)
    yield
    BaseContainer._resolvers = state


@fixture(scope="function")
def common_resolvers(restore_resolvers: Any) -> Any:
    """
    A fixture to register the common `identity` resolver.
    It depends on `restore_resolvers` to make it easier and safer to use.
    """

    def cast(t: Any, v: Any) -> Any:
        return {"str": str, "int": int}[t](v)  # cast `v` to type `t`

    OmegaConf.register_new_resolver("cast", cast)
    OmegaConf.register_new_resolver("identity", lambda x: x)

    yield



import copy
import sys
from typing import Any, Dict, List, MutableMapping, MutableSequence, Tuple, Union

from pytest import mark, param, raises

from omegaconf import (
    MISSING,
    DictConfig,
    ListConfig,
    OmegaConf,
    ReadonlyConfigError,
    ValidationError,
)
from omegaconf._utils import is_structured_config
from omegaconf.errors import ConfigKeyError, UnsupportedValueType
from omegaconf.nodes import IntegerNode
from tests import (
    A,
    B,
    C,
    ConcretePlugin,
    ConfWithMissingDict,
    Dataframe,
    Group,
    IllegalType,
    InterpolationDict,
    InterpolationList,
    MissingDict,
    MissingList,
    OptTuple,
    Package,
    Plugin,
    User,
    Users,
)


@mark.parametrize(
    ("merge_function", "input_unchanged"),
    [
        param(OmegaConf.merge, True, id="merge"),
        param(OmegaConf.unsafe_merge, False, id="unsafe_merge"),
    ],
)
@mark.parametrize(
    "inputs, expected",
    [
        # dictionaries
        param([{}, {"a": 1}], {"a": 1}, id="dict"),
        param([{"a": None}, {"b": None}], {"a": None, "b": None}, id="dict:none"),
        param([{"a": 1}, {"b": 2}], {"a": 1, "b": 2}, id="dict"),
        param(
            [
                {"a": {"a1": 1, "a2": 2}},
                {"a": {"a1": 2}},
            ],
            {"a": {"a1": 2, "a2": 2}},
            id="dict",
        ),
        param([{"a": 1, "b": 2}, {"b": 3}], {"a": 1, "b": 3}, id="dict"),
        param(
            ({"a": 1}, {"a": {"b": 3}}), {"a": {"b": 3}}, id="dict:merge_dict_into_int"
        ),
        param(({"b": {"c": 1}}, {"b": 1}), {"b": 1}, id="dict:merge_int_dict"),
        param(({"list": [1, 2, 3]}, {"list": [4, 5, 6]}), {"list": [4, 5, 6]}),
        param(({"a": 1}, {"a": IntegerNode(10)}), {"a": 10}),
        param(({"a": 1}, {"a": IntegerNode(10)}), {"a": IntegerNode(10)}),
        param(({"a": IntegerNode(10)}, {"a": 1}), {"a": 1}),
        param(({"a": IntegerNode(10)}, {"a": 1}), {"a": IntegerNode(1)}),
        param(({"a": "???"}, {"a": {}}), {"a": {}}, id="dict_merge_into_missing"),
        param(
            ({"a": "???"}, {"a": {"b": 10}}),
            {"a": {"b": 10}},
            id="dict_merge_into_missing",
        ),
        param(
            ({"a": {"b": 10}}, {"a": "???"}),
            {"a": {"b": 10}},
            id="dict_merge_missing_onto",
        ),
        param(
            ({"a": {"b": 10}}, {"a": DictConfig(content="???")}),
            {"a": {"b": 10}},
            id="dict_merge_missing_onto",
        ),
        param(
            ({}, {"a": "???"}),
            {"a": "???"},
            id="dict_merge_missing_onto_no_node",
        ),
        param(
            (
                {"a": 0, "b": 1},
                {"a": "${b}", "b": "???"},
            ),
            {"a": "${b}", "b": 1},
            id="dict_merge_inter_to_missing",
        ),
        param(
            (
                {"a": [0], "b": [1]},
                {"a": ListConfig(content="${b}"), "b": "???"},
            ),
            {"a": ListConfig(content="${b}"), "b": [1]},
            id="dict_with_list_merge_inter_to_missing",
        ),
        # lists
        (([1, 2, 3], [4, 5, 6]), [4, 5, 6]),
        (([[1, 2, 3]], [[4, 5, 6]]), [[4, 5, 6]]),
        (([1, 2, {"a": 10}], [4, 5, {"b": 20}]), [4, 5, {"b": 20}]),
        param(({"a": "???"}, {"a": []}), {"a": []}, id="list_merge_into_missing"),
        param(
            ({"a": "???"}, {"a": [1, 2, 3]}),
            {"a": [1, 2, 3]},
            id="list_merge_into_missing",
        ),
        param(
            ({"a": [1, 2, 3]}, {"a": "???"}),
            {"a": [1, 2, 3]},
            id="list_merge_missing_onto",
        ),
        param(
            ([1, 2, 3], ListConfig(content=MISSING)),
            ListConfig(content=[1, 2, 3]),
            id="list_merge_missing_onto",
        ),
        param(
            ({"a": 10, "list": []}, {"list": ["${a}"]}),
            {"a": 10, "list": [10]},
            id="merge_list_with_interpolation",
        ),
        # Interpolations
        # value interpolation
        param(
            ({"d1": 1, "inter": "${d1}"}, {"d1": 2}),
            {"d1": 2, "inter": 2},
            id="inter:updating_data",
        ),
        param(
            ({"d1": 1, "d2": 2, "inter": "${d1}"}, {"inter": "${d2}"}),
            {"d1": 1, "d2": 2, "inter": 2},
            id="inter:value_inter_over_value_inter",
        ),
        param(
            ({"inter": "${d1}"}, {"inter": 123}),
            {"inter": 123},
            id="inter:data_over_value_inter",
        ),
        param(
            ({"inter": "${d1}", "d1": 1, "n1": {"foo": "bar"}}, {"inter": "${n1}"}),
            {"inter": {"foo": "bar"}, "d1": 1, "n1": {"foo": "bar"}},
            id="inter:node_inter_over_value_inter",
        ),
        param(
            ({"inter": 123}, {"inter": "${data}"}),
            {"inter": "${data}"},
            id="inter:inter_over_data",
        ),
        # node interpolation
        param(
            ({"n": {"a": 10}, "i": "${n}"}, {"n": {"a": 20}}),
            {"n": {"a": 20}, "i": {"a": 20}},
            id="node_inter:node_update",
        ),
        param(
            ({"d": 20, "n": {"a": 10}, "i": "${n}"}, {"i": "${d}"}),
            {"d": 20, "n": {"a": 10}, "i": 20},
            id="node_inter:value_inter_over_node_inter",
        ),
        param(
            ({"n": {"a": 10}, "i": "${n}"}, {"i": 30}),
            {"n": {"a": 10}, "i": 30},
            id="node_inter:data_over_node_inter",
        ),
        param(
            ({"n1": {"a": 10}, "n2": {"b": 20}, "i": "${n1}"}, {"i": "${n2}"}),
            {"n1": {"a": 10}, "n2": {"b": 20}, "i": {"b": 20}},
            id="node_inter:node_inter_over_node_inter",
        ),
        param(
            ({"v": 10, "n": {"a": 20}}, {"v": "${n}"}),
            {"v": {"a": 20}, "n": {"a": 20}},
            id="inter:node_inter_over_data",
        ),
        param(
            (
                {"n": {"a": 10}, "i": "${n}"},
                {"i": {"b": 20}},
            ),
            {"n": {"a": 10}, "i": {"a": 10, "b": 20}},
            id="inter:node_over_node_interpolation",
        ),
        # Structured configs
        (({"user": User}, {}), {"user": User(name=MISSING, age=MISSING)}),
        (({"user": User}, {"user": {}}), {"user": User(name=MISSING, age=MISSING)}),
        (
            ({"user": User}, {"user": {"name": "Joe"}}),
            {"user": User(name="Joe", age=MISSING)},
        ),
        (
            ({"user": User}, {"user": {"name": "Joe", "age": 10}}),
            {"user": User(name="Joe", age=10)},
        ),
        ([{"users": Users}], {"users": {"name2user": {}}}),
        ([Users], {"name2user": {}}),
        ([Users, {"name2user": {}}], {"name2user": {}}),
        (
            [Users, {"name2user": {"joe": User}}],
            {"name2user": {"joe": {"name": MISSING, "age": MISSING}}},
        ),
        (
            [Users, {"name2user": {"joe": User(name="joe")}}],
            {"name2user": {"joe": {"name": "joe", "age": MISSING}}},
        ),
        param(
            [Users, {"name2user": {"joe": {"name": "joe"}}}],
            {"name2user": {"joe": {"name": "joe", "age": MISSING}}},
            id="users_merge_with_missing_age",
        ),
        param(
            [ConfWithMissingDict, {"dict": {"foo": "bar"}}],
            {"dict": {"foo": "bar"}},
            id="conf_missing_dict",
        ),
        param(
            [{}, ConfWithMissingDict],
            {"dict": "???"},
            id="merge_missing_dict_into_missing_dict",
        ),
        param(
            [{"user": User}, {"user": Group}],
            raises(ValidationError),
            id="merge_group_onto_user_error",
        ),
        param(
            [Plugin, ConcretePlugin], ConcretePlugin, id="merge_subclass_on_superclass"
        ),
        param(
            [{"user": "???"}, {"user": Group}],
            {"user": Group},
            id="merge_into_missing_node",
        ),
        param(
            [{"admin": {"name": "joe", "age": 42}}, Group(admin=None)],
            {"admin": None},
            id="merge_none_into_existing_node",
        ),
        param(
            [{"user": User()}, {"user": {"foo": "bar"}}],
            raises(ConfigKeyError),
            id="merge_unknown_key_into_structured_node",
        ),
        param(
            [OptTuple, {"x": [1, 2]}],
            {"x": [1, 2]},
            id="merge_list_into_optional_tuple_none",
        ),
        # DictConfig with element_type of Structured Config
        param(
            (
                DictConfig({}, element_type=User),
                {"user007": {"age": 99}},
            ),
            {"user007": {"name": "???", "age": 99}},
            id="dict:merge_into_sc_element_type:expanding_new_element",
        ),
        param(
            (
                DictConfig({"user007": "???"}, element_type=User),
                {"user007": {"age": 99}},
            ),
            {"user007": {"name": "???", "age": 99}},
            id="dict:merge_into_sc_element_type:into_missing_element",
        ),
        param(
            (
                DictConfig({"user007": User("bond", 7)}, element_type=User),
                {"user007": {"age": 99}},
            ),
            {"user007": {"name": "bond", "age": 99}},
            id="dict:merge_into_sc_element_type:merging_with_existing_element",
        ),
        param(
            (
                DictConfig({"user007": None}, element_type=User),
                {"user007": {"age": 99}},
            ),
            {"user007": {"name": "???", "age": 99}},
            id="dict:merge_into_sc_element_type:merging_into_none",
        ),
        # missing DictConfig
        param(
            [{"dict": DictConfig(content="???")}, {"dict": {"foo": "bar"}}],
            {"dict": {"foo": "bar"}},
            id="merge_into_missing_DictConfig",
        ),
        # missing Dict[str, str]
        param(
            [MissingDict, {"dict": {"foo": "bar"}}],
            {"dict": {"foo": "bar"}},
            id="merge_into_missing_Dict[str,str]",
        ),
        # missing ListConfig
        param(
            [{"list": ListConfig(content="???")}, {"list": [1, 2, 3]}],
            {"list": [1, 2, 3]},
            id="merge_into_missing_ListConfig",
        ),
        # missing List[str]
        param(
            [MissingList, {"list": ["a", "b", "c"]}],
            {"list": ["a", "b", "c"]},
            id="merge_into_missing_List[str]",
        ),
        # merging compatible dict into MISSING structured config expands it
        # to ensure the resulting node follows the protocol set by the underlying type
        param([B, {"x": {}}], {"x": {"a": 10}}, id="structured_merge_into_missing"),
        param(
            [B, {"x": {"a": 20}}], {"x": {"a": 20}}, id="structured_merge_into_missing"
        ),
        param([C, {"x": A}], {"x": {"a": 10}}, id="structured_merge_into_none"),
        param([C, C], {"x": None}, id="none_not_expanding"),
        # Merge into list with Structured Config
        param(
            [ListConfig(content=[], element_type=User), [{}]],
            [User()],
            id="list_sc_element_merge_dict",
        ),
        param(
            [ListConfig(content=[], element_type=User), [{"name": "Bond", "age": 7}]],
            [User(name="Bond", age=7)],
            id="list_sc_element_merge_dict",
        ),
        param(
            [ListConfig(content=[], element_type=User), [{"name": "Bond"}]],
            [User(name="Bond", age=MISSING)],
            id="list_sc_element_merge_dict",
        ),
    ],
)
def test_merge(
    inputs: Any,
    expected: Any,
    merge_function: Any,
    input_unchanged: bool,
) -> None:
    configs = [OmegaConf.create(c) for c in inputs]

    if isinstance(expected, (MutableMapping, MutableSequence)) or is_structured_config(
        expected
    ):
        merged = merge_function(*configs)
        assert merged == expected
        if input_unchanged:
            # test input configs are not changed.
            # Note that converting to container without resolving to avoid resolution errors while comparing
            for i in range(len(inputs)):
                input_i = OmegaConf.create(inputs[i])
                orig = OmegaConf.to_container(input_i, resolve=False)
                merged2 = OmegaConf.to_container(configs[i], resolve=False)
                assert orig == merged2
    else:
        with expected:
            merge_function(*configs)


def test_merge_error_retains_type() -> None:
    cfg = OmegaConf.structured(ConcretePlugin)
    with raises(ValidationError):
        cfg.merge_with({"params": {"foo": "error"}})
    assert OmegaConf.get_type(cfg) == ConcretePlugin


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
def test_primitive_dicts(merge: Any) -> None:
    c1 = {"a": 10}
    c2 = {"b": 20}
    merged = merge(c1, c2)
    assert merged == {"a": 10, "b": 20}


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
@mark.parametrize("a_, b_, expected", [((1, 2, 3), (4, 5, 6), [4, 5, 6])])
def test_merge_no_eq_verify(
    merge: Any, a_: Tuple[int], b_: Tuple[int], expected: Tuple[int]
) -> None:
    a = OmegaConf.create(a_)
    b = OmegaConf.create(b_)
    c = merge(a, b)
    # verify merge result is expected
    assert expected == c


@mark.parametrize(
    "c1,c2,expected",
    [({}, {"a": 1, "b": 2}, {"a": 1, "b": 2}), ({"a": 1}, {"b": 2}, {"a": 1, "b": 2})],
)
def test_merge_with(c1: Any, c2: Any, expected: Any) -> None:
    a = OmegaConf.create(c1)
    b = OmegaConf.create(c2)
    a.merge_with(b)
    assert a == expected


@mark.parametrize(
    "c1,c2,expected",
    [({}, {"a": 1, "b": 2}, {"a": 1, "b": 2}), ({"a": 1}, {"b": 2}, {"a": 1, "b": 2})],
)
def test_merge_with_c2_readonly(c1: Any, c2: Any, expected: Any) -> None:
    a = OmegaConf.create(c1)
    b = OmegaConf.create(c2)
    OmegaConf.set_readonly(b, True)
    a.merge_with(b)
    assert a == expected
    assert OmegaConf.is_readonly(a)


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
def test_3way_dict_merge(merge: Any) -> None:
    c1 = OmegaConf.create("{a: 1, b: 2}")
    c2 = OmegaConf.create("{b: 3}")
    c3 = OmegaConf.create("{a: 2, c: 3}")
    c4 = merge(c1, c2, c3)
    assert {"a": 2, "b": 3, "c": 3} == c4


def test_merge_list_list() -> None:
    a = OmegaConf.create([1, 2, 3])
    b = OmegaConf.create([4, 5, 6])
    a.merge_with(b)
    assert a == b


@mark.parametrize("merge_func", [OmegaConf.merge, OmegaConf.unsafe_merge])
@mark.parametrize(
    "base, merge, exception",
    [
        ({}, [], TypeError),
        ([], {}, TypeError),
        ([1, 2, 3], None, ValueError),
        ({"a": 10}, None, ValueError),
        (Package, {"modules": [{"foo": "var"}]}, ConfigKeyError),
    ],
)
def test_merge_error(merge_func: Any, base: Any, merge: Any, exception: Any) -> None:
    base = OmegaConf.create(base)
    merge = None if merge is None else OmegaConf.create(merge)
    with raises(exception):
        merge_func(base, merge)


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
@mark.parametrize(
    "c1, c2",
    [
        param({"foo": "bar"}, {"zoo": "foo"}, id="dict"),
        param([1, 2, 3], [4, 5, 6], id="list"),
    ],
)
def test_with_readonly_c1(merge: Any, c1: Any, c2: Any) -> None:
    cfg1 = OmegaConf.create(c1)
    cfg2 = OmegaConf.create(c2)
    OmegaConf.set_readonly(cfg1, True)
    cfg3 = merge(cfg1, cfg2)
    assert OmegaConf.is_readonly(cfg3)


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
@mark.parametrize(
    "c1, c2",
    [
        param({"foo": "bar"}, {"zoo": "foo"}, id="dict"),
        param([1, 2, 3], [4, 5, 6], id="list"),
    ],
)
def test_with_readonly_c2(merge: Any, c1: Any, c2: Any) -> None:
    cfg1 = OmegaConf.create(c1)
    cfg2 = OmegaConf.create(c1)
    OmegaConf.set_readonly(cfg2, True)
    cfg3 = merge(cfg1, cfg2)
    assert OmegaConf.is_readonly(cfg3)


@mark.parametrize("c1, c2", [({"foo": "bar"}, {"zoo": "foo"}), ([1, 2, 3], [4, 5, 6])])
def test_into_readonly(c1: Any, c2: Any) -> None:
    cfg = OmegaConf.create(c1)
    OmegaConf.set_readonly(cfg, True)
    with raises(ReadonlyConfigError):
        cfg.merge_with(c2)


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
@mark.parametrize(
    "c1, c2, expected",
    [
        (
            {"node": {"foo": "bar"}},
            {"node": {"zoo": "foo"}},
            {"node": {"foo": "bar", "zoo": "foo"}},
        ),
    ],
)
def test_dict_merge_readonly_into_readwrite(
    merge: Any, c1: Any, c2: Any, expected: Any
) -> None:
    c1 = OmegaConf.create(c1)
    c2 = OmegaConf.create(c2)
    OmegaConf.set_readonly(c2.node, True)
    with raises(ReadonlyConfigError):
        c2.node.foo = 10
    assert OmegaConf.merge(c1, c2) == expected
    c1.merge_with(c2)
    assert c1 == expected


@mark.parametrize(
    "c1, c2, expected",
    [({"node": [1, 2, 3]}, {"node": [4, 5, 6]}, {"node": [4, 5, 6]})],
)
def test_list_merge_readonly_into_readwrite(c1: Any, c2: Any, expected: Any) -> None:
    c1 = OmegaConf.create(c1)
    c2 = OmegaConf.create(c2)
    OmegaConf.set_readonly(c2.node, True)
    with raises(ReadonlyConfigError):
        c2.node.append(10)
    assert OmegaConf.merge(c1, c2) == expected
    c1.merge_with(c2)
    assert c1 == expected


def test_parent_maintained() -> None:
    c1 = OmegaConf.create({"a": {"b": 10}})
    c2 = OmegaConf.create({"aa": {"bb": 100}})
    # only test OmegaConf.merge. unsafe_merge is failing this test by design
    c3 = OmegaConf.merge(c1, c2)
    assert isinstance(c1, DictConfig)
    assert isinstance(c2, DictConfig)
    assert isinstance(c3, DictConfig)
    assert c1.a._get_parent() is c1
    assert c2.aa._get_parent() is c2
    assert c3.a._get_parent() is c3


@mark.parametrize(
    "cfg,overrides,expected",
    [
        ([1, 2, 3], ["0=bar", "2.a=100"], ["bar", 2, dict(a=100)]),
        ({}, ["foo=bar", "bar=100"], {"foo": "bar", "bar": 100}),
        ({}, ["foo=bar=10"], {"foo": "bar=10"}),
    ],
)
def test_merge_with_dotlist(
    cfg: Union[List[Any], Dict[str, Any]],
    overrides: List[str],
    expected: Union[List[Any], Dict[str, Any]],
) -> None:
    c = OmegaConf.create(cfg)
    c.merge_with_dotlist(overrides)
    assert c == expected


def test_merge_with_cli() -> None:
    c = OmegaConf.create([1, 2, 3])
    sys.argv = ["app.py", "0=bar", "2.a=100"]
    c.merge_with_cli()
    assert c == ["bar", 2, dict(a=100)]


@mark.parametrize(
    "dotlist, expected",
    [([], {}), (["foo=1"], {"foo": 1}), (["foo=1", "bar"], {"foo": 1, "bar": None})],
)
def test_merge_empty_with_dotlist(dotlist: List[str], expected: Dict[str, Any]) -> None:
    c = OmegaConf.create()
    c.merge_with_dotlist(dotlist)
    assert c == expected


@mark.parametrize("dotlist", ["foo=10", ["foo=1", 10]])
def test_merge_with_dotlist_errors(dotlist: List[str]) -> None:
    c = OmegaConf.create()
    with raises(ValueError):
        c.merge_with_dotlist(dotlist)


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
def test_merge_allow_objects(merge: Any) -> None:
    iv = IllegalType()
    cfg = OmegaConf.create({"a": 10})
    with raises(UnsupportedValueType):
        merge(cfg, {"foo": iv})

    cfg._set_flag("allow_objects", True)
    ret = merge(cfg, {"foo": iv})
    assert ret == {"a": 10, "foo": iv}


def test_merge_with_allow_Dataframe() -> None:
    cfg = OmegaConf.create({"a": Dataframe()}, flags={"allow_objects": True})
    ret = OmegaConf.merge({}, cfg)
    assert isinstance(ret.a, Dataframe)


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
@mark.parametrize(
    "dst, other, expected, node",
    [
        param(
            OmegaConf.structured(InterpolationList),
            OmegaConf.create({"list": [0.1]}),
            {"list": [0.1]},
            "list",
            id="merge_interpolation_list_with_list",
        ),
        param(
            OmegaConf.structured(InterpolationDict),
            OmegaConf.create({"dict": {"a": 4}}),
            {"dict": {"a": 4}},
            "dict",
            id="merge_interpolation_dict_with_dict",
        ),
    ],
)
def test_merge_with_src_as_interpolation(
    merge: Any, dst: Any, other: Any, expected: Any, node: Any
) -> None:
    res = merge(dst, other)
    assert res == expected


@mark.parametrize("merge", [OmegaConf.merge, OmegaConf.unsafe_merge])
@mark.parametrize(
    "dst, other, node",
    [
        param(
            OmegaConf.structured(InterpolationDict),
            OmegaConf.structured(InterpolationDict),
            "dict",
            id="merge_interpolation_dict_with_interpolation_dict",
        ),
        param(
            OmegaConf.structured(InterpolationList),
            OmegaConf.structured(InterpolationList),
            "list",
            id="merge_interpolation_list_with_interpolation_list",
        ),
    ],
)
def test_merge_with_other_as_interpolation(
    merge: Any, dst: Any, other: Any, node: Any
) -> None:
    res = merge(dst, other)
    assert OmegaConf.is_interpolation(res, node)


@mark.parametrize(
    ("c1", "c2"),
    [
        param(
            ListConfig(content=[1, 2, 3], element_type=int),
            ["a", "b", "c"],
            id="merge_with_list",
        ),
    ],
)
def test_merge_with_error_not_changing_target(c1: Any, c2: Any) -> Any:
    backup = copy.deepcopy(c1)
    with raises(ValidationError):
        c1.merge_with(c2)
    assert c1 == backup


@mark.parametrize(
    "register_func",
    [OmegaConf.legacy_register_resolver, OmegaConf.register_new_resolver],
)
def test_into_custom_resolver_that_throws(
    restore_resolvers: Any, register_func: Any
) -> None:
    def fail() -> None:
        raise ValueError()

    register_func("fail", fail)

    configs = (
        {"d": 20, "i": "${fail:}"},
        {"i": "zzz"},
    )
    expected = {"d": 20, "i": "zzz"}
    assert OmegaConf.merge(*configs) == expected



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

from pytest import raises

from omegaconf import (
    MISSING,
    DictConfig,
    MissingMandatoryValue,
    OmegaConf,
    ReadonlyConfigError,
    ValidationError,
)


class Height(Enum):
    SHORT = 0
    TALL = 1


@dataclass
class SimpleTypes:
    num: int = 10
    pi: float = 3.1415
    is_awesome: bool = True
    height: Height = Height.SHORT
    description: str = "text"


def test_simple_types_class() -> None:
    # Instantiate from a class
    conf = OmegaConf.structured(SimpleTypes)
    assert conf.num == 10
    assert conf.pi == 3.1415
    assert conf.is_awesome is True
    assert conf.height == Height.SHORT
    assert conf.description == "text"


def test_static_typing() -> None:
    conf: SimpleTypes = OmegaConf.structured(SimpleTypes)
    assert conf.description == "text"  # passes static type checking
    with raises(AttributeError):
        # This will fail both the static type checking and at runtime
        # noinspection PyStatementEffect
        conf.no_such_attribute  # type: ignore


def test_simple_types_obj() -> None:
    # Instantiate from an Object, any value can be overridden
    # at construction
    conf = OmegaConf.structured(SimpleTypes(num=20, pi=3))
    assert conf.num == 20
    assert conf.pi == 3
    # Everything not overridden at construction takes the default value
    assert conf.is_awesome is True
    assert conf.height == Height.SHORT
    assert conf.description == "text"


def test_conversions() -> None:
    conf: SimpleTypes = OmegaConf.structured(SimpleTypes)

    # OmegaConf can convert types at runtime
    conf.num = 20  # ok, type matches

    # ok, the String "20" is converted to the int 20
    conf.num = "20"  # type: ignore

    assert conf.num == 20
    with raises(ValidationError):
        # ValidationError: "one" cannot be converted to an integer
        conf.num = "one"  # type: ignore

    # booleans can take many forms
    for expected, values in {
        True: ["on", "yes", "true", True, "1"],
        False: ["off", "no", "false", False, "0"],
    }.items():
        for b in values:
            conf.is_awesome = b  # type: ignore
            assert conf.is_awesome == expected

    # Enums too
    for expected1, values1 in {
        Height.SHORT: [Height.SHORT, "Height.SHORT", "SHORT", 0],
        Height.TALL: [Height.TALL, "Height.TALL", "TALL", 1],
    }.items():
        for b in values1:
            conf.height = b  # type: ignore
            assert conf.height == expected1


@dataclass
class Modifiers:
    # regular field
    num: int = 10

    # Fields can be optional
    optional_num: Optional[int] = None

    # MISSING fields must be populated at runtime before access. accessing them while they
    # are missing will result in a MissingMandatoryValue exception
    another_num: int = MISSING


def test_modifiers() -> None:
    conf: Modifiers = OmegaConf.structured(Modifiers)
    # regular fields cannot take None
    with raises(ValidationError):
        conf.num = None  # type: ignore

    # but Optional fields can
    conf.optional_num = None
    assert conf.optional_num is None

    # Accessing a missing field will trigger MissingMandatoryValue exception
    with raises(MissingMandatoryValue):
        # noinspection PyStatementEffect
        conf.another_num

    # but you can access it once it's been assigned
    conf.another_num = 42
    assert conf.another_num == 42


@dataclass
class User:
    # A simple user class with two missing fields
    name: str = MISSING
    height: Height = MISSING


# Group class contains two instances of User.
@dataclass
class Group:
    name: str = MISSING
    # data classes can be nested
    admin: User = User  # type: ignore

    # You can also specify different defaults for nested classes
    manager: User = User(name="manager", height=Height.TALL)


def test_nesting() -> None:
    conf = OmegaConf.structured(Group)
    assert conf == {
        "name": "???",
        "admin": {"name": MISSING, "height": MISSING},
        "manager": {"name": "manager", "height": Height.TALL},
    }

    expected = """name: ???
admin:
  name: ???
  height: ???
manager:
  name: manager
  height: TALL
"""
    assert OmegaConf.to_yaml(conf) == expected

    # you can assign a different object of the same type
    conf.admin = User(name="omry", height=Height.TALL)
    with raises(ValidationError):
        # but not incompatible types
        conf.admin = 10

    with raises(ValidationError):
        # You can't assign a dict even if the field matches
        conf.manager = {"name": "secret", "height": Height.TALL}


@dataclass
class Lists:
    # List with Any as type can take any primitive type OmegaConf supports:
    # int, float, bool, str and Enums as well as Any (which is the same as not having a specified type).
    untyped_list: List[Any] = field(default_factory=lambda: [1, "foo", True])

    # typed lists can hold int, bool, str, float or enums.
    int_list: List[int] = field(default_factory=lambda: [10, 20, 30])


def test_typed_list_runtime_validation() -> None:
    conf = OmegaConf.structured(Lists)

    conf.untyped_list[0] = True  # okay, list can hold any primitive type

    conf.int_list[0] = 999  # okay
    assert conf.int_list[0] == 999

    conf.int_list[0] = "1000"  # also ok!
    assert conf.int_list[0] == 1000

    with raises(ValidationError):
        conf.int_list[0] = "fail"


# Dicts
@dataclass
class Dicts:
    # Key must be a string or Enum, value can be any primitive type OmegaConf supports:
    # int, float, bool, str and Enums as well as Any (which is the same as not having a specified type).
    untyped_dict: Dict[str, Any] = field(
        default_factory=lambda: {"foo": True, "bar": 100}
    )

    # maps string to height Enum
    str_to_height: Dict[str, Height] = field(
        default_factory=lambda: {"Yoda": Height.SHORT, "3-CPO": Height.TALL}
    )


def test_typed_dict_runtime_validation() -> None:
    conf = OmegaConf.structured(Dicts)
    conf.untyped_dict["foo"] = "buzz"  # okay, list can hold any primitive type
    conf.str_to_height["Shorty"] = Height.SHORT  # Okay
    with raises(ValidationError):
        # runtime failure, cannot convert True to Height
        conf.str_to_height["Yoda"] = True


# Frozen
@dataclass(frozen=True)
class FrozenClass:
    x: int = 10
    list: List[int] = field(default_factory=lambda: [1, 2, 3])


def test_frozen() -> None:
    # frozen classes are read only, attempts to modify them at runtime
    # will result in a ReadonlyConfigError
    conf = OmegaConf.structured(FrozenClass)
    with raises(ReadonlyConfigError):
        conf.x = 20

    # Read-only flag is recursive
    with raises(ReadonlyConfigError):
        conf.list[0] = 20


class Protocol(Enum):
    HTTP = 0
    HTTPS = 1


@dataclass
class Domain:
    name: str = MISSING
    path: str = MISSING
    protocols: List[Protocol] = field(default_factory=lambda: [Protocol.HTTPS])


@dataclass
class WebServer(DictConfig):
    protocol_ports: Dict[Protocol, int] = field(
        default_factory=lambda: {Protocol.HTTP: 80, Protocol.HTTPS: 443}
    )
    # Dict of name to domain
    domains: Dict[str, Domain] = field(default_factory=dict)

    # List of all domains
    domains_list: List[Domain] = field(default_factory=list)


# Test that Enum can be used a dictionary key
def test_enum_key() -> None:
    conf = OmegaConf.structured(WebServer)
    # When an Enum is a dictionary key the name of the Enum is actually used
    # as the key
    assert conf.protocol_ports.HTTP == 80
    assert conf.protocol_ports["HTTP"] == 80
    assert conf.protocol_ports[Protocol.HTTP] == 80


def test_dict_of_objects() -> None:
    conf: WebServer = OmegaConf.structured(WebServer)
    conf.domains["blog"] = Domain(name="blog.example.com", path="/www/blog.example.com")
    with raises(ValidationError):
        conf.domains.foo = 10  # type: ignore

    assert conf.domains["blog"].name == "blog.example.com"
    assert conf.domains["blog"].path == "/www/blog.example.com"
    assert conf == {
        "protocol_ports": {Protocol.HTTP: 80, Protocol.HTTPS: 443},
        "domains": {
            "blog": {
                "name": "blog.example.com",
                "path": "/www/blog.example.com",
                "protocols": [Protocol.HTTPS],
            }
        },
        "domains_list": [],
    }


def test_list_of_objects() -> None:
    conf: WebServer = OmegaConf.structured(WebServer)
    conf.domains_list.append(
        Domain(name="blog.example.com", path="/www/blog.example.com")
    )
    with raises(ValidationError):
        conf.domains_list.append(10)  # type: ignore

    assert conf.domains_list[0].name == "blog.example.com"
    assert conf.domains_list[0].path == "/www/blog.example.com"
    assert conf == {
        "protocol_ports": {Protocol.HTTP: 80, Protocol.HTTPS: 443},
        "domains": {},
        "domains_list": [
            {
                "name": "blog.example.com",
                "path": "/www/blog.example.com",
                "protocols": [Protocol.HTTPS],
            }
        ],
    }


def test_merge() -> None:
    @dataclass
    class Config:
        num: int = 10
        user: User = User(name=MISSING, height=MISSING)
        domains: Dict[str, Domain] = field(default_factory=dict)

    yaml = """
user:
    name: Omry
domains:
    blog_website:
        name: blog
        protocols:
          - HTTPS
"""

    schema: Config = OmegaConf.structured(Config)
    cfg = OmegaConf.create(yaml)
    merged: Any = OmegaConf.merge(schema, cfg)
    assert merged == {
        "num": 10,
        "user": {"name": "Omry", "height": "???"},
        "domains": {
            "blog_website": {
                "name": "blog",
                "path": "???",
                "protocols": [Protocol.HTTPS],
            }
        },
    }
    assert OmegaConf.is_missing(merged.domains.blog_website, "path")


def test_merge_example() -> None:
    @dataclass
    class Server:
        port: int = MISSING

    @dataclass
    class Log:
        file: str = MISSING
        rotation: int = MISSING

    @dataclass
    class MyConfig:
        server: Server = Server()
        log: Log = Log()
        users: List[str] = field(default_factory=list)
        numbers: List[int] = field(default_factory=list)

    schema = OmegaConf.structured(MyConfig)
    with raises(ValidationError):
        OmegaConf.merge(schema, OmegaConf.create({"log": {"rotation": "foo"}}))

    with raises(ValidationError):
        cfg = schema.copy()
        cfg.numbers.append("fo")

    with raises(ValidationError):
        OmegaConf.merge(schema, OmegaConf.create({"numbers": ["foo"]}))



import re
from typing import Any, Optional

from pytest import mark, param, raises

from omegaconf import OmegaConf
from omegaconf.errors import InterpolationResolutionError


@mark.parametrize(
    "value,expected",
    [
        # We only test a few typical cases: more extensive grammar tests are
        # found in `test_grammar.py`.
        # bool
        ("false", False),
        ("true", True),
        # int
        ("10", 10),
        ("-10", -10),
        # float
        ("10.0", 10.0),
        ("-10.0", -10.0),
        # null
        ("null", None),
        ("NulL", None),
        # strings
        ("hello", "hello"),
        ("hello world", "hello world"),
        ("  123  ", "  123  "),
        ('"123"', "123"),
        # lists and dicts
        ("[1, 2, 3]", [1, 2, 3]),
        ("{a: 0, b: 1}", {"a": 0, "b": 1}),
        ("[\t1, 2, 3\t]", [1, 2, 3]),
        ("{   a: b\t  }", {"a": "b"}),
        # interpolations
        ("${parent.sibling}", 1),
        ("${.sibling}", 1),
        ("${..parent.sibling}", 1),
        ("${uncle}", 2),
        ("${..uncle}", 2),
        ("${oc.env:MYKEY}", 456),
    ],
)
def test_decode(monkeypatch: Any, value: Optional[str], expected: Any) -> None:
    monkeypatch.setenv("MYKEY", "456")
    c = OmegaConf.create(
        {
            # The node of interest is "node" (others are used to test interpolations).
            "parent": {
                "node": f"${{oc.decode:'{value}'}}",
                "sibling": 1,
            },
            "uncle": 2,
        }
    )
    assert c.parent.node == expected


def test_decode_none() -> None:
    c = OmegaConf.create({"x": "${oc.decode:null}"})
    assert c.x is None


@mark.parametrize(
    ("value", "exc"),
    [
        param(
            123,
            raises(
                InterpolationResolutionError,
                match=re.escape(
                    "TypeError raised while resolving interpolation: "
                    "`oc.decode` can only take strings or None as input, but `123` is of type int"
                ),
            ),
            id="bad_type",
        ),
        param(
            "'[1, '",
            raises(
                InterpolationResolutionError,
                match=re.escape(
                    "GrammarParseError raised while resolving interpolation: "
                    "missing BRACKET_CLOSE at '<EOF>'"
                ),
            ),
            id="parse_error",
        ),
        param(
            # Must be escaped to prevent resolution before feeding it to `oc.decode`.
            "'\\${foo}'",
            raises(
                InterpolationResolutionError,
                match=re.escape("Interpolation key 'foo' not found"),
            ),
            id="interpolation_not_found",
        ),
    ],
)
def test_decode_error(monkeypatch: Any, value: Any, exc: Any) -> None:
    c = OmegaConf.create({"x": f"${{oc.decode:{value}}}"})
    with exc:
        c.x
